from pydantic import BaseModel


class RelocationResource(BaseModel):
    column_name: str
    """Name of the column in the input data representing the resource."""
    amount: float
    """Amount of the resource needed to serve one unit of demand."""




from collections import defaultdict
import numpy as np
from .adalina_zoning_data import AdalinaZoningData
import time
import logging


class AdalinaZoningSolution:

    def __init__(self, data : AdalinaZoningData):
        self.data = data
        self.gdf_export = None

        self.labels = dict()
        self.weights = defaultdict(float)
        self.intraclust_distance = defaultdict(float)

        self.representatives = dict()
        self._K = -1

        self.cl_dict_set = None
        self.cl_dict_list = None
        self.infeasibilities = None
        self.is_feasible = False

        self.unconnected_clusters = []

        self.constr_heur_starttime = None

    def _check_timelim(self, timelimit):

        return time.time() - self.constr_heur_starttime <= timelimit

    def constructive(self, timelimit = 60, fout_log =  None):

        self.constr_heur_starttime = time.time()

        try:

            # get Kmax representative maximizing average distance?
            R = []
            WR = []
            n = self.data.node_pairs_costs_df.to_numpy()
            iv = n.max(axis=0)
            i = iv.argmax()
            j = n[i].argmax()

            i = int(i)
            R.append(i)
            WR.append(self.data.get_weight_node(i))
            j = int(j)
            R.append(j)
            WR.append(self.data.get_weight_node(j))

            while len(R) < self.data.Kmax and self._check_timelim(timelimit):
                v = n[:, R].mean(axis=1)

                m = None
                mi = None
                for i, el in enumerate(v):
                    if i in R:
                        continue
                    if m is None or el > m:
                        m = el
                        mi = i

                if mi is None:
                    break

                R.append(mi)
                WR.append(self.data.get_weight_node(mi))

            if not self._check_timelim(timelimit):
                return False

            R = np.array(R)
            WR = np.array(WR)
            R = R[np.argsort(WR)[::-1]]
            WR = np.sort(WR)[::-1]

            A = []
            k = 0
            N = list(self.data.G.nodes.keys())
            C = defaultdict(list)
            all_weights = dict()
            for n in N:
                if n in R:
                    continue
                all_weights[n]=self.data.get_weight_node(n)

            # self.data.Fmin = 3

            for j, r in enumerate(R):
                if len(C) >= self.data.Kmin or np.setdiff1d(N, R.tolist() + A).shape[0] == 0:
                    break

                wr = WR[j]
                C[r].append(r)

                if wr >= self.data.Fmin:
                    continue
                else:
                    i = 0
                    while i < len(C[r]) and self._check_timelim(timelimit):
                        r_el = C[r][i]
                        i += 1
                        #while wr < self.data.Fmin and np.setdiff1d(N, R.tolist() + A).shape[0] > 0:
                        wmax = []
                        wmaxi = []

                        for nk in self.data.G.neighbors(r_el): #, nw in all_weights.items():
                            if nk in A or nk in R:
                                continue
                            wmax.append(all_weights[nk])
                            wmaxi.append(nk)

                        if len(wmax) == 0:
                            continue

                        # wmax = np.array(wmax)
                        wmaxi = np.array(wmaxi)
                        wmaxi = wmaxi[np.argsort(wmax)[::-1]]
                        wmax = np.sort(wmax)[::-1]

                        for k, el in enumerate(wmax):
                            if wr + wmax[:(k+1)].sum() >= self.data.Fmin:
                                break

                        for l in range(k+1):
                            A.append(wmaxi[l])
                            C[r].append(wmaxi[l])
                            wr += wmax[l]

                        if wr >= self.data.Fmin:
                            break

                    if wr < self.data.Fmin:
                        for el in C[r]:
                            if el in A:
                                A.remove(el)
                            elif el in R:
                                i_el = np.where(R == el)[0][0]
                                R = np.concatenate((R[:i_el], R[(i_el+1):]))
                                # R.remove(el)
                        del(C[r])

            if len(C) < self.data.Kmin:
                return False

            if not self._check_timelim(timelimit):
                return False

            # aggiungi nodi mancanti ai cluster presenti, uno alla volta
            # aggiungendo per primo il nodo adiacente a un cluster con distanza minore
            # 1 - libera candidati rappresentanti non in C
            R = list(C.keys()) # np.setdiff1d(R, list(C.keys()))
            # 2 -
            while len(A) + len(R) < self.data.V and self._check_timelim(timelimit):

                dval = defaultdict(list)
                vval = defaultdict(list)
                for k, el in C.items():
                    vicini_esterni = {
                        v for u in el for v in self.data.G.neighbors(u)
                        if v not in el and v not in R and v not in A
                    }
                    for v in vicini_esterni:
                        val = self.data.node_pairs_costs_df.loc[v, el].sum()
                        dval[k].append(val)
                        vval[k].append(v)

                    if len(dval[k]) == 0:
                        continue

                    # dval[k] = np.array(dval[k])
                    vval[k] = np.array(vval[k])[np.argsort(dval[k])]
                    # dval[k] = np.sort(dval[k])

                    C[k].append(vval[k][0])
                    A.append(vval[k][0])

                # C[minvc].append(minvi)
                # A.append(minvi)

            if not self._check_timelim(timelimit):
                return False

            for k, el in C.items():
                for i, n in enumerate(el):
                    for j in range(i+1, len(el)):
                        n2 = el[j]
                        self.add_edge(n, n2)
                self.add_repr(k)

            return self.check_feasibility()

        except: # Exception as e:
            logging.error("constructive heuristic failed!")

        return False

    @property
    def labels_(self):
        _l = []
        for i, k in enumerate(self.data.G.nodes):
            _l.append(self.labels[k])
        return _l

    def get_K(self):
        return self._K + 1

    def add_belonging(self, i, k):

        if i in self.labels and self.labels[i] != k:
            raise ValueError("")

        self.labels[i] = k
        self.intraclust_distance[k] += self.data.get_cost_edge(i, k)

    def check_nodes_belong_same_cluster(self, nodes: list):

        nodes_k = [self.labels[n]  for n in nodes if n in self.labels]
        k = nodes_k[0]
        return all([el == k for el in nodes_k])

    def add_edge(self, i, j):

        k = -1
        if i in self.labels:
            k = self.labels[i]
        if j in self.labels:
            if k != -1 and k != self.labels[j]:
                raise ValueError(f"ERROR: nodi {i} e {j} già presenti in cluster diversi: {k}, {self.labels[j]}")

            k = self.labels[j]

        if k == -1:
            self._K += 1
            k = self._K

        self.labels[i] = k
        self.labels[j] = k
        self.intraclust_distance[k] += self.data.get_cost_edge(i,j)

    def add_repr(self, i, self_index=False):
        if i not in self.labels:
            self._K += 1
            i_index = self._K
            if self_index:
                i_index = i
            self.labels[i] = i_index

        self.representatives[i] = self.labels[i]

    def __repr__(self):
        top = f"{self.get_K()} cluster \n"

        if self.get_K() <= 0:
            return top

        if self.cl_dict_set is None:
            self.build_cluster_representations()

        top += "{ \n"
        for ck, cv in self.cl_dict_list.items():
            top += f" {ck} : ["
            top += ','.join([str(el) for el in list(cv)])
            top += '], \n'
        top += "}\n"

        if self.infeasibilities is not None and len(self.infeasibilities) > 0:
            top += "INFEASIBILITIES: \n"
            for el in self.infeasibilities:
                top += el + "\n"

        return top

    def to_json(self):

        # cl = self.get_clustering()

        j = dict()
        j['clusters'] = []

        for k, v in self.cl_dict_list.items():
            j['clusters'].append({
                'representative' : k,
                "nodes" : v,
                "distance" : self.intraclust_distance[k],
                "weight" : self.weights[k]
            }
        )

        return j

    # def get_clustering(self):
    #     cl = self.build_cluster_representations()
    #     ser_cl = None
    #     if cl is not None:
    #         ser_cl = dict()
    #         for k, v in self.cl_dict_set.items():
    #             ser_cl[k] = list(v)
    #
    #         self.data.gdf = self.data.gdf.add_clustering(ser_cl)
    #
    #     return ser_cl

    def build_cluster_representations(self):

        if self.cl_dict_set is not None:
            return # self.cl_dict_set

        self.cl_dict_set = dict()
        self.cl_dict_list = dict()

        for rk, rv in self.representatives.items():
            self.cl_dict_set[rv] = set()
            self.cl_dict_set[rv].add(rk)

        for lk, lv in self.labels.items():
            self.cl_dict_set[lv].add(lk)
            self.weights[lv] += self.data.get_weight_node(lk)

        for k, val in self.cl_dict_set.items():
            self.cl_dict_list[k] = list(val)

        self.gdf_export = self.data.add_clustering(self.labels)

        # return self.cl_dict_set

    def get_dataframe_to_export(self):
        return self.gdf_export

    def check_feasibility(self):

        self.build_cluster_representations()

        self.unconnected_clusters = []

        self.infeasibilities = []
        self.is_feasible = True

        missing_nodes = [ k for k in self.data.G.nodes if k not in self.labels ]
        if len(missing_nodes) > 0:
            self.is_feasible = False
            self.infeasibilities.append(
                f"nodi non assegnati a cluster : {missing_nodes}"
            )

        if self.get_K() < self.data.Kmin:
            self.infeasibilities.append(
                f"numero di cluster {self.get_K()} minore del lower bound {self.data.Kmin}"
            )
            self.is_feasible = False

        if self.get_K() > self.data.Kmax:
            self.infeasibilities.append(
                f"numero di cluster {self.get_K()} maggiore dell'upper bound {self.data.Kmax}"
            )
            self.is_feasible = False

        for ck, cv in self.cl_dict_set.items():
            if not self.data.check_nodes_subset_connected(list(cv)):
                self.infeasibilities.append(
                    f"cluster non connesso [" + ','.join([str(el) for el in list(cv)]) +  "]"
                )
                self.unconnected_clusters.append(list(cv))
                self.is_feasible = False

            peso_minimo = 0
            for el in cv:
                peso_minimo += self.data.get_weight_node(el)
            if peso_minimo < self.data.Fmin:
                self.infeasibilities.append(
                    f"cluster con peso totale {peso_minimo} minore del consentito {self.data.Fmin}"
                )
                self.is_feasible = False

        return self.is_feasible


from .algorithms.adalina_relocation_algorithm import relocation_algorithm
from .models.relocation_resource import RelocationResource

__all__ = [
    'relocation_algorithm',
    'RelocationResource'
]


import networkx as nx
from libadalina_core.sedona_utils import DataFrame, to_spark_dataframe, DEFAULT_EPSG
import pyspark.sql as ps
from pyspark import Row
from pyspark.sql import functions as func
from pyspark.sql.functions import udf
from pyspark.sql.types import StructType, StringType, NumericType, IntegerType, LongType
from sedona.sql import ST_Intersects, ST_Distance, ST_Centroid, ST_SetSRID, ST_SRID

@udf(returnType=LongType())
def first_item_udf(value: Row):
    if value:
        return value._1
    return -1


def get_shape_node(
        graph: nx.Graph,
        shape_df: DataFrame,
        index_column: str,
        columns_to_keep: list[str] = None
) -> ps.DataFrame:
    """
    """
    shape_df = to_spark_dataframe(shape_df)

    # Convert nodes to spark DataFrame
    edges_df = (shape_df.sparkSession.createDataFrame(
        (((p1, p2), data['geometry']) for p1, p2, data in graph.edges(data=True)),
        ['edge', 'geometry']
    ))
    edges_df = edges_df.withColumn('geometry', ST_SetSRID(edges_df.geometry, DEFAULT_EPSG.value))

    shape_df = (shape_df
            .join(edges_df, on=ST_Intersects(shape_df.geometry, edges_df.geometry), how='inner')
            .select('*', ST_Distance(ST_Centroid(shape_df.geometry), edges_df.geometry).alias('distance_to_centroid'))
            .groupBy(shape_df[index_column])
                .agg(
                    *(func.first(c).alias(c) for c in (shape_df.columns if columns_to_keep is None else columns_to_keep) if c != 'geometry'),
                    func.min_by(edges_df.edge, 'distance_to_centroid').alias('closest_edge')
                )
            .select('*', first_item_udf(func.col('closest_edge')).alias('closest_node'))
            )

    if columns_to_keep is not None and 'geometry' not in columns_to_keep:
        shape_df = shape_df.drop('geometry')
    return shape_df



from libadalina_analytics.clustering.models.adalina_zoning_data import AdalinaZoningData
from libadalina_analytics.clustering.models.adalina_zoning_solution import AdalinaZoningSolution

import highspy
import numpy as np
import time
import logging
from sklearn.cluster import AgglomerativeClustering

hscb = highspy.cb

# model.callbacks
# highspy.highs.HighsCallback

"""
    sum_{(s,t) in E : s in S, t in T x_{st} >= x_{ij} 
    forall cut [S,T] : i in S, j in T
"""

# def lazy_SEC_constraint(callback_type : hscb.HighsCallbackType,
#                   message : str,
#                   data_out : hscb.HighsCallbackDataOut,
#                   data_in : hscb.HighsCallbackDataOut,
#                   user_callback_data : object) -> None:
#
#     if callback_type == hscb.HighsCallbackType.kCallbackMipSolution:
#         print("callback mip_solution")
#         print(data_out.mip_solution.to_array(len(user_callback_data.vardict)))
#
#         # print("obj. fun callback ", data_out.objective_function_value)
#         # print("obj. fun callback ", data_out.running_time)
#
#         assert isinstance(user_callback_data, AdalinaZoningModel)
#
#         # if not user_callback_data._build_solution():
#         #    if len(user_callback_data.solution.unconnected_clusters) > 0:
#
#                 # costruisci gomory-hu su G con i valori delle x
#                 # per ogni cluster disconnesso:
#                 # per ogni coppia di nodi
#
#         #        for cl in user_callback_data.solution.unconnected_clusters:
#         #            print(cl)
#
#         pass
#
#     # elif callback_type == hscb.kCallbackMipDefineLazyConstraints:
#     #    print("LAZY CONSTRAINT")

# VAR
# x_ij {1 if i in cluster with representative j, 0 otherwise}
# y_j {1 if j is representative of a cluster, 0 otherwise}

# Kmin <= sum{j in N} y_j <= Kmax
# sum{j in N} x_ij + y_i = 1 forall i in N
# x_{ij} <= sum{k in N : k in adj(i) and dist^-(k, j) <= dist^-(i,j) } forall (i,j) : i not in adj(j)
# x_{ik} + x_{jk} <= 1 + w_{ij} forall i,j,k in N^3 : i \not= j \not= k
# x_{ik} + w_{ik} <= 1 forall (i,k) in N^2 : i \not= k
# min \sum{(i,j) \in N^2 : i \not= j} c[i,j] * (x_{ij} + w_{ij})
# \sum_{i in N :  i\not=j } x_ij * c_i >= y_j * F_min forall j in N

class AdalinaZoningModel:

    def __init__(self, data : AdalinaZoningData,  log_fout = None):

        self.log_fout = log_fout
        self.data = data
        self.vardict = dict()
        self.model = highspy.Highs()
        inf = highspy.kHighsInf
        self.run_status = None
        self.solution = None
        self.best_feasible_sol = None

        xvar = []
        for (i,j) in data.E1:
            xvar.append(
                self._add_variable(0, 1,
                    data.get_cost_edge(i,j),
                    highspy.HighsVarType.kInteger,
               f"x_{i}_{j}")
            )

        # y[s] in {0,1}
        zvar = []
        firstnode = True
        for i in data.G.nodes:
            if firstnode:
                zvar.append(
                    self._add_variable(1,1, 0,
                                       highspy.HighsVarType.kInteger,
                                     f"z_{i}")
                                 )
                firstnode = False
            else:
                zvar.append(
                    self._add_variable(0,1, 0,
                                       highspy.HighsVarType.kInteger,
                                     f"z_{i}")
                                 )

        # x_{ij} + x_{jk} - x_{ik} \le 1  & \forall i, j, k \in V \colon (i,j), (j,k), (i,k) \in E'
        for i in data.G.nodes:
            for j in data.edges[i]:
                for k in data.edges[j]:
                    if k <= i:
                        continue

                    for el in range(3):
                        _varvalues = [1]*3
                        _varvalues[el] = -1
                        self.model.addRow(-inf, 1, 3,
                                          np.array([
                                              self._get_vardict_index(f'x_{i}_{j}'),
                                              self._get_vardict_index(f'x_{j}_{k}'),
                                              self._get_vardict_index(f'x_{i}_{k}')
                                          ]),
                                          np.array(_varvalues)
                                          )

        # \sum_{(i,j) \in E'} w_j x_{ij}  \ge \left(F^{\min} - w_i\right) z_i & \forall i \in V
        for i in data.G.nodes:
            _vars = [self._get_vardict_index(f'z_{i}')]
            _varvalues = [- self.data.Fmin + self.data.get_weight_node(i)]
            for j in data.edges[i]:
                _vars.append(self._get_vardict_index(f'x_{i}_{j}'))
                _varvalues.append(self.data.get_weight_node(j))

            self.model.addRow(0, +inf, len(_vars),
                              np.array(_vars),
                              np.array(_varvalues)
                              )

        # \sum_{j \in V \colon j <= i} x_{ji} \le |V| (1-z_i) & \forall i \in V
        for i in data.G.nodes:
            _vars = [self._get_vardict_index(f'z_{i}')]
            _varvalues = [self.data.V]
            for j in data.G.nodes:
                if j >=i:
                    continue

                _vars.append(self._get_vardict_index(f'x_{j}_{i}'))
                _varvalues.append(1)

            self.model.addRow(-inf, self.data.V,
                              len(_vars),
                              np.array(_vars),
                              np.array(_varvalues)
                              )

        # \sum_{j\in V \colon j <= i} x_{ji} + z_i \ge 1 & \forall i \in V
        for i in data.G.nodes:
            _vars = [self._get_vardict_index(f'z_{i}')]
            _varvalues = [1]
            for j in data.G.nodes:
                if j >=i:
                    continue

                _vars.append(self._get_vardict_index(f'x_{j}_{i}'))
                _varvalues.append(1)

            self.model.addRow(1, inf,
                              len(_vars),
                              np.array(_vars),
                              np.array(_varvalues)
                              )

        # K^{\min} \le \sum_{i\in V} z_i \le K^{\max}
        _vars = [self._get_vardict_index(f'z_{i}') for i in data.G.nodes]
        _varvalues = np.ones(self.data.V)
        self.model.addRow(self.data.Kmin, self.data.Kmax,
                          len(_vars),
                          np.array(_vars),
                          np.array(_varvalues)
                          )

        # self.model.setCallback(lazy_SEC_constraint, self)
        # self.model.startCallback(hscb.HighsCallbackType.kCallbackMipSolution)
        # self.model.startCallback(hscb.HighsCallbackType.kCallbackMipDefineLazyConstraints)


    def _add_variable_dict_element(self, name):
        self.vardict[name] = len(self.vardict)

    def _add_variable(self, lb, ub, obj, vartype, name):
        el = self.model.addVariable(lb, ub,
                                   obj = obj,
                                   type=vartype,
                                   name=name)
        self._add_variable_dict_element(name)

        return el

    def _get_vardict_index(self, name):
        return self.vardict[name]

    def run(self, timelimit=60, label_sol = None):

        self.solution = None
        self.model.setOptionValue("time_limit",
            #highspy.HighsOptions.time_limit,
                                  timelimit/5)

        if label_sol is None:
            try:
                agglom_clust = AgglomerativeClustering(linkage='average',
                                                metric='precomputed',
                                                n_clusters=self.data.Kmax).fit(self.data.node_pairs_costs_df)

                label_sol = agglom_clust.labels_
            except:
                logging.debug("AgglomaritveClustering failed!", self.log_fout)
                pass

        if label_sol is not None:
            xwarmstart = []
            for el in np.unique(label_sol):
                areas = np.where(label_sol==el)[0]
                xwarmstart.append(self._get_vardict_index(f'z_{areas[0]}'))
                for i, _el in enumerate(areas):
                    for j in range(i+1, areas.shape[0]):
                        xwarmstart.append(self._get_vardict_index(f'x_{_el}_{areas[j]}'))

            xwarmstart = np.array(xwarmstart, dtype=int)
            self.model.setSolution(xwarmstart.shape[0], xwarmstart, np.ones(xwarmstart.shape[0]))

        start_time = time.time()

        while True:
            self.solution = None
            self.run_status = self.model.run()

            self.get_solution(run_model=False)

            if self.solution is None:
                break

            if time.time() - start_time > timelimit:
                break

            if not self.solution.is_feasible and len(self.solution.unconnected_clusters) > 0:
                # costruisci gomory-hu su G con i valori delle x
                # per ogni cluster disconnesso:
                # per ogni coppia di nodi

                xvarvaluedict = dict()
                allvars = self.model.getVariables()

                for e in self.data.G.edges:
                    xvarvaluedict[e] = self.model.variableValue(allvars[self._get_vardict_index(f'x_{e[0]}_{e[1]}')])

                self.data.gomory_hu(xvarvaluedict)

                for cl in self.solution.unconnected_clusters:
                    # for u, v in zip(cl, cl[1:]):
                    for i, u in enumerate(cl):
                        for j in range(i+1, len(cl)):
                            v = cl[j]

                            if (u, v) in self.data.G.edges:
                                continue

                            _, (S, T) = self.data.minimum_edge_weight_in_shortest_path(u, v)

                            # sum_{(s,t) in E : s in S, t in T x_{st} >= x_{ij}
                            #     forall cut [S,T] : i in S, j in T
                            _vars = []
                            _varvalues = [-1]

                            if u < v:
                                _vars.append(self._get_vardict_index(f'x_{u}_{v}'))
                            elif u > v:
                                _vars.append(self._get_vardict_index(f'x_{v}_{u}'))

                            for s in S:
                                for t in T:
                                    if (s,t) in self.data.G.edges:
                                        if s < t:
                                            _vars.append(self._get_vardict_index(f'x_{s}_{t}'))
                                        else:
                                            _vars.append(self._get_vardict_index(f'x_{t}_{s}'))
                                        _varvalues.append(1)

                            # print_log("adding new cut" )
                            self.model.addRow(0, +highspy.kHighsInf,
                                              len(_vars),
                                              np.array(_vars),
                                              np.array(_varvalues)
                                              )

                # self.model.writeModel("/home/marco/Desktop/prova.lp")
            else:
                break

        return self.run_status

    def _build_solution(self):
        self.solution = AdalinaZoningSolution(self.data)
        allvars = self.model.getVariables()

        _repr = []
        for varname, varindex in self.vardict.items():

            val = self.model.variableValue(allvars[varindex])
            if val < 1e-5:
                continue

            l = varname.split("_")

            if l[0].startswith('x'):
                self.solution.add_edge(int(l[1]), int(l[2]))
            elif l[0].startswith('z'):
                _repr.append(int(l[1]))

        for r in _repr:
            self.solution.add_repr(r)

        return self.solution.check_feasibility()

    def get_solution(self, run_model=True):

        if self.solution is not None:
            return self.solution

        info = self.model.getInfo()
        if (self.run_status == highspy.HighsStatus.kOk or
              (self.run_status == highspy.HighsStatus.kWarning and
               info.primal_solution_status == highspy.SolutionStatus.kSolutionStatusFeasible)):

            logging.debug(f"obj. fun. value {self.model.getObjectiveValue()}")

            if self._build_solution():
                logging.debug("solution feasible!")

            else:
                logging.debug("model solved but solution infeasible!")

            return self.solution

        elif self.run_status == highspy.HighsStatus.kError:

            logging.debug("no solution available. model ended with Error")
            return None

        if self.solution is None and run_model:

            self.run()
            return self.get_solution(run_model=False)

        logging.debug("no solution available. run model with function run()")
        return None



import argparse
from .adalina_model_type import AdalinaModelType

class AdalinaAlgorithmOptions:

    def __init__(self):

        self.basedir = ""
        self.outdir = ""
        self.fileprefix = ""
        self.modeltype_str = None
        self.export_LPfile = False
        self.timelimit = 60
        self.run_hierarchical = False
        self.modeltype = None
        self.log_fout = None
        self.run_parametric_analysis = False

    @classmethod
    def from_argparse(cls, options : argparse.Namespace):

        obj = cls()

        obj.basedir = options.basedir
        obj.outdir = options.outdir
        obj.fileprefix = options.fileprefix

        if hasattr(options, "modeltype_str"):
            obj.modeltype_str = options.modeltype_str

        if hasattr(options, "export_LPfile"):
            obj.export_LPfile = options.export_LPfile

        if hasattr(options, "timelimit"):
            obj.timelimit = options.timelimit

        if obj.modeltype_str is not None:
            if "HIERARCHICAL" not in obj.modeltype_str :
                obj.modeltype = AdalinaModelType.from_label(obj.modeltype_str)
            else:
                obj.run_hierarchical = True
                if "_PA" in obj.modeltype_str:
                    obj.run_parametric_analysis = True
                obj.modeltype = AdalinaModelType.MINSUM_UNSERVED
        else:
            obj.run_hierarchical = True
            obj.modeltype = AdalinaModelType.MINSUM_UNSERVED

        return obj

    def change_model_type(self, newmodeltype):
        assert isinstance(newmodeltype, AdalinaModelType)
        self.modeltype = newmodeltype
        self.modeltype_str = newmodeltype.get_label()



from libadalina_core.sedona_utils import DataFrame, EPSGFormats
from libadalina_analytics.utils import GeometryFormats
from .adalina_algorithms import run_hierarchy_with_distance_threshold
from ..models import AdalinaData, AdalinaSolution, AdalinaAlgorithmOptions
from ..models.adalina_solution import get_solution_csv_AMELIA
import pandas as pd

from ..models.relocation_resource import RelocationResource


def relocation_algorithm(data: DataFrame,
                         epsg: EPSGFormats,
                         id_column: str = 'id',
                         geometry_column: str = 'geometry',
                         demand_column: str = 'demand',
                         geometry_format: GeometryFormats = GeometryFormats.WKT,
                         max_distance_assignment: float | None = None,
                         max_distance_relocation: float | None = None,
                         server_column: str | None = None,
                         resources: list[RelocationResource] | None = None,
                         timelimit: int = 60) -> pd.DataFrame | None:
    """
    Optimally assign demand to servers minimizing relocation costs.

    Parameters
    ----------
    data : pandas.DataFrame
        The input data containing the geometries and properties.
    epsg : EPSGFormats
        The EPSG format of the input geometries.
    id_column : str
        The name of the column containing the unique identifiers for each location. Default is 'id'.
    geometry_column : str
        The name of the column containing the geometries. Default is 'geometry'.
    demand_column : str
        The name of the column containing the demand values. Default is 'demand'.
    geometry_format : GeometryFormats
        The format of the geometries in the geometry column. Default is GeometryFormats.WKT.
    max_distance_assignment : float | None
        The maximum distance allowed for assigning demand to servers. If None, no limit is applied. Default is None.
    max_distance_relocation : float | None
        The maximum distance allowed for relocating demands. If None, no limit is applied. Default is None.
    server_column : str | None
        The name of the column indicating whether a location is a server (True) or not (False).
        If None, all locations are considered as potential servers. Default is None.
    resources : list[RelocationResource] | None
        A list of RelocationResource objects representing the resources available at each server.
        Each RelocationResource contains:
            - column_name: The name of the column in the input data representing the resource
            - amount: The total amount of the resource needed to serve one unit of demand.
    timelimit : int
        The maximum time (in seconds) to run the algorithm. Default is 60 seconds.

    Returns
    -------
    pandas.DataFrame | None
        A DataFrame containing the optimal assignment of demand to servers, or None if no solution is found.
    """
    user_input: dict = {
        "epsg": epsg,
        "geometry_type": geometry_format,
        "geometry": geometry_column,
        "timelimit": timelimit
    }
    if id_column is not None:
        user_input["IDs"] = id_column
    if max_distance_assignment is not None:
        user_input["max_distance_assignment"] = max_distance_assignment
    if max_distance_relocation is not None:
        user_input["max_distance_relocation"] = max_distance_relocation
    if demand_column is not None:
        user_input["demand"] = demand_column
    if resources is not None:
        user_input["resources"] = [{
            'name': resource.column_name,
            'amount': resource.amount
        } for resource in resources]
    if server_column is not None:
        user_input["is_server"] = server_column

    data = AdalinaData.from_Amelia(amelia_file=data,
                                   user_input=user_input
                                   )

    # 4 - RUN ADALINA HIERARCHICAL ALGORITHM
    options = AdalinaAlgorithmOptions()  # I keep the default options
    _, all_solutions = run_hierarchy_with_distance_threshold(data, options)

    if len(all_solutions) == 0:
        return None

    return get_solution_csv_AMELIA(all_solutions[-1], data, epsg, geometry_format)



import argparse

class AdalinaZoningAlgorithmOptions:

    def __init__(self):

        self.basedir = ""
        self.outdir = ""
        self.fileprefix = ""
        self.modeltype_str = ""
        self.export_LPfile = ""
        self.timelimit = ""
        self.log_fout = None

    @classmethod
    def from_argparse(cls,
                      options : argparse.Namespace):

        obj = cls()

        obj.basedir = options.basedir
        obj.outdir = options.outdir
        obj.fileprefix = options.fileprefix

        obj.modeltype_str = None
        if hasattr(options, "modeltype_str"):
            obj.modeltype_str = options.modeltype_str

        obj.export_LPfile = False
        if hasattr(options, "export_LPfile"):
            obj.export_LPfile = options.export_LPfile

        obj.timelimit = 60
        if hasattr(options, "timelimit"):
            obj.timelimit = options.timelimit

        return obj


import json
import pandas as pd
import shapely

from libadalina_core.sedona_utils.coordinate_formats import EPSGFormats
from libadalina_analytics.utils import GeometryFormats
from .adalina_data import AdalinaData, EPSG_FORMAT_FOR_ALGORITHM
from .adalina_model_type import AdalinaModelType


class AdalinaSolution:

    def __init__(self, objfun_value=None):

        self.assignments = None
        self.all_relocations = None
        self.total_usage = None
        self.objfun_value = objfun_value

        self.xvarvalues = pd.DataFrame(columns=['shelter', 'area', 'value'])
        self.xvarvalues = self.xvarvalues.set_index(['shelter', 'area'])

        self.rhovalues = pd.DataFrame(columns=["shelter_from", "shelter_to", "resource", "value"])
        self.rhovalues = self.rhovalues.set_index(["shelter_from", "shelter_to", "resource"])

        self.yvarvalues = pd.DataFrame(columns=['shelter', 'value'])
        self.yvarvalues = self.yvarvalues.set_index(['shelter'])

        self.wvarvalues = pd.DataFrame(columns=['area', 'value'])
        self.wvarvalues = self.wvarvalues.set_index(['area'])

        self.zvarvalues = pd.DataFrame(columns=['shelter', 'resource', 'value'])
        self.zvarvalues = self.zvarvalues.set_index(['shelter', 'resource'])

        self.usagevarvalues = pd.DataFrame(columns=['shelter', 'value'])
        self.usagevarvalues = self.usagevarvalues.set_index(['shelter'])

        self.wmaxvalue = None
        self.rhomaxvalue = None
        self.w2varvalue = None

        self.feasible = False

        self.areas = pd.DataFrame(columns=['area', 'unserved', 'unserved_fract',
                                           'cost_transport', 'cost_unserved'])

        self.areas = self.areas.set_index(['area'])

        self.shelter = pd.DataFrame(columns=['shelter', 'isopen', 'cost_transport',
                                             'cost_open','cost_excess', 'fract_usage_kpi'])
        self.shelter = self.shelter.set_index(['shelter'])

        self.relocation = pd.DataFrame(columns = ['shelter', 'resource',
                                                  'need', 'exiting', 'entering', 'excess'])
        # self.relocation = self.relocation.set_index(['shelter', 'resource'])

        self.total_cost_unserved = 0
        self.total_cost_excess = 0
        self.total_cost_transport = 0
        self.total_cost_open = 0
        self.objfun_value_recomputed = 0
        self.total_cost_relocation = 0
        self.total_cost_unserved_w2 = 0

        self.infeasibility_log = []

        self.modeltype = None
        self.ub_unserved = None
        self.ub_assign_distance = None
        self.ub_max_usage = None

    def export_json(self, outfilename):

        serializable_dict = {
            "feasible" : self.feasible,
             "areas" : self.areas.to_dict(),
             "shelters" : self.shelter.to_dict(),
             # "relocation" : self.relocation.groupby(level=0).apply(
             #  lambda df: df.xs(df.name).to_dict()['value']).to_dict(),
             "cost_assignments" : float(self.total_cost_transport),
             "cost_open_facility": float(self.total_cost_open),
             "cost_excess_capacity": float(self.total_cost_excess),
             "cost_unserved": float(self.total_cost_unserved),
             "obj_fun_value" : float(self.objfun_value)
        }

        if self.modeltype is not None:

            assert isinstance(self.modeltype, AdalinaModelType)

            serializable_dict["modeltype"] = self.modeltype.get_label()

        if self.ub_unserved is not None:

            serializable_dict["UB_unserved"] = self.ub_unserved

        if self.ub_assign_distance is not None:
            serializable_dict["UB_assign_distance"] = self.ub_assign_distance

        if self.wmaxvalue is not None:
            serializable_dict["wmax"] = float(self.wmaxvalue)

        if self.rhomaxvalue is not None:
            serializable_dict["rhomax"] = float(self.rhomaxvalue)

        if self.xvarvalues.shape[0] > 0:
            serializable_dict["xvar"] = self.xvarvalues.groupby(level=0).apply(
                lambda df: df.xs(df.name).to_dict()['value']).to_dict()
        if self.yvarvalues.shape[0] > 0:
            serializable_dict["yvar"] = self.yvarvalues.to_dict()['value']
        if self.wvarvalues.shape[0] > 0:
            serializable_dict["wvar"] = self.wvarvalues.to_dict()['value']

        if self.zvarvalues.shape[0] > 0:
            serializable_dict["zvar"] = self.zvarvalues.groupby(level=0).apply(
                lambda df: df.xs(df.name).to_dict()['value']).to_dict()

        if self.rhovalues.shape[0] > 0:
            serializable_dict["rhovar"] = {}
            for sfrom in self.rhovalues.index.unique(0):
                rhofrom = self.rhovalues.xs(sfrom, level=0)
                serializable_dict["rhovar"][sfrom] = {}

                for sto in rhofrom.index.unique(0):
                    rhofromto = rhofrom.xs(sto, level=0)

                    serializable_dict["rhovar"][sfrom][sto] = rhofromto.to_dict()['value']

                # self.rhovalues.groupby(level=0).apply(
                # lambda df: df.xs(df.name).to_dict()['value']).to_dict()

        # Dump the serializable dictionary as a JSON file
        with open(outfilename+'.json', 'w') as fout:
            json.dump(serializable_dict, fout)

        self.assignments.to_csv(outfilename+"_assignments.csv", sep=";", index=None)
        self.all_relocations.to_csv(outfilename+"_relocations.csv", sep=";", index=None)

    def check_wmax_set(self):
        return self.wmaxvalue is not None

    def check_rhomax_set(self ):
        return self.rhomaxvalue is not None

    def set_wmaxvalue(self, value):
        self.wmaxvalue = value

    def set_rhomaxvalue(self, value):
        self.rhomaxvalue =  value

    def add_xvarvalue(self, shelter, area, value):
        self.xvarvalues.loc[(shelter, area), 'value'] = value

    def add_yvarvalue(self, shelter, value):
        self.yvarvalues.loc[shelter, 'value'] = value

    def add_wvarvalue(self, area, value):
        self.wvarvalues.loc[area, 'value'] = value

    def add_usagevarvalue(self, shelter, value):
        self.usagevarvalues.loc[shelter, 'value'] = value

    def add_zvarvalue(self, shelter, resource, value):
        self.zvarvalues.loc[ (shelter, resource), 'value'] = value

    def add_rhovalue(self, shelter_from, shelter_to, resource, value):
        self.rhovalues.loc[(shelter_from, shelter_to, resource), "value"] = value

    def add_w2varvalue(self, area, shelter, value):

        if self.w2varvalue is None:
            self.w2varvalue = pd.DataFrame(columns=['shelter', 'area', 'value'])
            self.w2varvalue = self.w2varvalue.set_index(['shelter', 'area'])
        self.w2varvalue.loc[(shelter, area), 'value'] = value

    def _set_infeasibility(self, logtext):
        self.feasible = False
        self.infeasibility_log.append(logtext)

    def check_feasibility(self, data, modeltype):

        assert isinstance(data, AdalinaData)
        assert isinstance(modeltype, AdalinaModelType)

        self.modeltype = modeltype

        self.assignments = self.xvarvalues.reset_index()

        for a in self.assignments.area.unique():
            self.assignments.loc[self.assignments.area == a, "value"] = self.assignments.loc[
                                                     self.assignments.area == a, "value"] * data.get_area_demand(a)

        self.all_relocations = self.rhovalues.reset_index()
        for s in self.all_relocations.shelter_from.unique():
            for r in self.all_relocations.loc[self.all_relocations.shelter_from == s].resource.unique():
                self.all_relocations.loc[(self.all_relocations.shelter_from == s) &
                                     (self.all_relocations.resource == r)
                                        , "value"] = \
                    self.all_relocations.loc[(self.all_relocations.shelter_from == s) &
                                         (self.all_relocations.resource == r)
                    , "value"]*data.get_resource_available(s, r)

        self.feasible = True

        for a in data.A:
            self.areas.loc[a] = 0.0
        for s in data.S:
            self.shelter.loc[s] = 0.0

            for r in data.R:
                self.relocation.loc[ self.relocation.shape[0], ["shelter", "resource"]] = [s,r]

        self.relocation = self.relocation.set_index(["shelter", "resource"])
        self.relocation[self.relocation.isna()] = 0

        self.shelter['isopen'] = self.shelter['isopen'].astype(bool)
        for s in self.yvarvalues.index:

            if not data.shelters.loc[s, 'usable']:
                self._set_infeasibility(f"shelter {s} opened but not usable")

            self.shelter.loc[s, 'isopen'] = True

        # sum[s] r[s,a]*x[s,a] + w[a] == 1 forall a
        for a in data.A:
            cover = 0

            _xvararea = self.xvarvalues[self.xvarvalues.index.isin([a], level=1)]

            for _, row in _xvararea.iterrows():

                s, _ = row.name

                val = row["value"]

                cover += val

                # x[s,a] <= r[a,s]
                if val > 1e-5 and not data.get_is_reachable(s, a):
                    self._set_infeasibility(f"assignment area {a} to shelter {s} while unreachable")

            if a in self.wvarvalues.index:

                val = self.wvarvalues.loc[a].iloc[0]
                cover +=  val

                self.areas.loc[a, 'unserved_fract'] += val

            if cover < 1 - 1e-5:
                self._set_infeasibility(f"area {a} not fully covered {cover}")

            self.areas.loc[a, 'unserved'] = self.areas.loc[a, 'unserved_fract']*data.get_area_demand(a)

        if modeltype in [AdalinaModelType.MINDIST_ASSIGNMENTS, AdalinaModelType.MINMAX_USAGEFAC] and \
            data.ub_total_unserved_demand is not None:
            demand_unserved = self.areas['unserved'].sum()

            self.ub_unserved = data.ub_total_unserved_demand

            if demand_unserved - data.ub_total_unserved_demand > 1e-5:
                self._set_infeasibility(f" total unserved demand {demand_unserved} exceeds UB {data.ub_total_unserved_demand}")

        # sum[a] d[a] * x[s,a] <= sum{s1 in S} rho[s1,s,r] + z[r, s] forall s in S, forall r
        for s in data.S:
            for r in data.R:

                need = 0
                entering = 0

                isopen = s in self.yvarvalues.index

                _xvarshelter = self.xvarvalues[self.xvarvalues.index.isin([s], level=0)]
                for _, row in _xvarshelter.iterrows():
                    _, area = row.name
                    need += data.get_area_demand(area) * row["value"] * data.get_resource_need(r)

                for s1 in data.S:
                    if (s1, s, r) in self.rhovalues.index:
                        val = self.rhovalues.loc[(s1, s, r), "value"]*data.get_resource_available(s1, r)
                        entering += val
                        self.relocation.loc[(s1,r), 'exiting'] += val

                        if not data.get_transport_possible(s, s1):
                            self._set_infeasibility(f"relocation between {s1} and {s} of {val} resources while not possible")

                if (s, r) in self.zvarvalues.index:
                    val = self.zvarvalues.loc[(s,r)].iloc[0]

                    # z[s] <= My[s]
                    if not open:
                        self._set_infeasibility(f"exceeding shelter {s} while not used")

                    # capacity += val
                    self.relocation.loc[(s,r), 'excess'] +=val

                if need > 1e-5 and not isopen:
                    self._set_infeasibility(f"using shelter {s} while not opened")

                self.relocation.loc[(s,r), 'need'] = need
                self.relocation.loc[(s, r), 'entering'] = entering

                excess = 0
                if (s,r) in self.zvarvalues.index:
                    excess = self.zvarvalues.loc[(s,r), "value"]
                self.relocation.loc[(s,r), 'excess'] = excess

                if need - entering - excess > 1e-2:
                    self._set_infeasibility(f"shelter {s} needs {need} of resource {r}, only having {entering+excess}")

        # objfun sum[a] c[a]d[a]w[a] + sum[a,s] c[s,a]*d[a]*x[s,a]  + sum[s] (co[s]y[s] + ce[s]z[s])

        # sum[a] c[a]d[a]w[a]
        for a in data.A:
            if a in self.wvarvalues.index:
                val = self.wvarvalues.loc[a].iloc[0]*data.get_demcost_notserved(a, modeltype) # data.get_area_demand(a)
                self.total_cost_unserved += val

                self.areas.loc[a, 'cost_unserved'] = val

        # sum[s] (co[s]y[s] + ce[s]z[r,s])
        for s in data.S:
            copen = data.get_fixed_cost_shelter(s, modeltype) # data.get_fixed_cost_shelter(s)
            if s in self.yvarvalues.index:
                self.shelter.loc[s, 'cost_open'] = copen

                self.total_cost_open += copen

                tot_res_used = self.relocation.xs(s, level=0).entering.sum()
                tot_res_init = sum([data.get_resource_available(s, r) for r in data.R])
                if tot_res_init > 1e-5:
                    self.shelter.loc[s, 'fract_usage_kpi'] = tot_res_used/tot_res_init

            for r in data.R:
                if (s,r) in self.zvarvalues.index:

                    val = self.zvarvalues.loc[(s,r), 'value']*data.get_cost_exceed_cap(s, modeltype) # data.get_cost_exceed_cap(s)

                    self.shelter.loc[s, 'cost_excess'] += val

                    self.relocation.loc[(s,r), 'excess'] = val

                    self.total_cost_excess += val

        for sfrom in data.S:
            for sto in data.S:
                if sfrom == sto:
                    continue
                for r in data.R:
                    if (sfrom, sto, r) in self.rhovalues.index:

                        rhoval = self.rhovalues.loc[(sfrom, sto, r), "value"]

                        self.total_cost_relocation += \
                            rhoval * \
                            data.get_cost_resource_transport(sfrom , sto, r, modeltype)

        # sum[a,s] c[s,a]*d[a]*x[s,a]
        for i, row in self.xvarvalues.iterrows():
            s, a = row.name
            val = row['value'] * data.get_cost_assign(s, a) * data.get_area_demand(a) #data.get_demcost_assign(s,a,modeltype) #

            self.total_cost_transport += val

            self.shelter.loc[s, 'cost_transport'] += val
            self.areas.loc[a, 'cost_transport'] += val

        if modeltype in [AdalinaModelType.MINMAX_USAGEFAC, AdalinaModelType.MINSUM_USAGEFAC] and data.ub_total_assignment_distance is not None:

            self.ub_assign_distance = data.ub_total_assignment_distance

            if self.total_cost_transport - data.ub_total_assignment_distance > 1e-5:
                self._set_infeasibility(f"total assignment cost {self.total_cost_transport} exceeds UB {data.ub_total_unserved_demand}")

        if modeltype == AdalinaModelType.MINSUM_USAGEFAC:

            self.total_usage = self.usagevarvalues.values.sum()

        if modeltype == AdalinaModelType.MAXSUM_DIST_UNSERVED:
            for i, row in self.w2varvalue.iterrows():
                s, a = row.name
                val = row['value'] * data.get_cost_assign(s, a) * data.get_area_demand(
                    a)  # data.get_demcost_assign(s,a,modeltype) #

                self.total_cost_unserved_w2 += val

        if modeltype == AdalinaModelType.MINSUM_UNSERVED:
            self.objfun_value_recomputed = self.total_cost_unserved
        elif modeltype == AdalinaModelType.MINMAX_UNSERVED:
            self.objfun_value_recomputed = self.wmaxvalue
        elif modeltype == AdalinaModelType.MINMAX_USAGEFAC:
            self.objfun_value_recomputed = self.rhomaxvalue
        elif modeltype == AdalinaModelType.MIN_EXTRARESOURCES:
            self.objfun_value_recomputed = self.total_cost_excess
        elif modeltype == AdalinaModelType.MINDIST_ASSIGNMENTS:
            self.objfun_value_recomputed = self.total_cost_transport
        elif modeltype == AdalinaModelType.MINSUM_USAGEFAC:
            self.objfun_value_recomputed = self.total_usage
        elif modeltype == AdalinaModelType.MINSUM_RELOCATION:
            self.objfun_value_recomputed = self.total_cost_relocation
        elif modeltype == AdalinaModelType.MAXSUM_DIST_UNSERVED:
            self.objfun_value_recomputed = self.total_cost_unserved_w2
        else:
            self.objfun_value_recomputed = self.total_cost_open  + self.total_cost_excess + self.total_cost_transport \
                                     + self.total_cost_unserved + self.total_cost_relocation

        if abs(self.objfun_value - self.objfun_value_recomputed) > 1e-2:
            self.infeasibility_log.append(f"objective function not matching recomputation: model {self.objfun_value} - recomputation {self.objfun_value_recomputed}")
            self.feasible = False

        return self.feasible


def get_solution_csv_AMELIA(sol: AdalinaSolution,
                            data : AdalinaData,
                            epsg: EPSGFormats = EPSGFormats.EPSG4326,
                            geometry_format: GeometryFormats = GeometryFormats.WKT
                            ):
    csv_out = pd.merge(data.areas.reset_index(), sol.areas.reset_index(), on = "area")

    csv_out["area"] = [el[5:] for el in csv_out["area"]]
    csv_out["area"] = csv_out["area"].astype(data.amelia_file_orig.dtypes[data.user_input["IDs"]])

    if epsg != EPSG_FORMAT_FOR_ALGORITHM:
        csv_out = csv_out.to_crs(epsg=epsg.value)

    if geometry_format == GeometryFormats.WKT:
        return csv_out.to_wkt()
    elif geometry_format == GeometryFormats.GEOJSON:
        output_copy = pd.DataFrame(csv_out.select_dtypes(exclude='geometry'))
        for col in csv_out.columns[csv_out.dtypes == "geometry"]:
            output_copy[col] = csv_out[col].apply(shapely.to_geojson)
        return output_copy

    return csv_out



from enum import Enum


class AdalinaModelType(Enum):

    MINSUM_ALLCOST = 0
    MIN_EXTRARESOURCES = 1
    MINSUM_UNSERVED = 2
    MINMAX_UNSERVED = 3
    MINMAX_USAGEFAC = 4
    MINDIST_ASSIGNMENTS = 5
    MINSUM_USAGEFAC = 6
    MINSUM_RELOCATION = 7
    MAXSUM_DIST_UNSERVED = 8

    @staticmethod
    def get_all_labels(get_str = True):
        _label_dict = {
            "ALLCOSTS": AdalinaModelType.MINSUM_ALLCOST,
            "MIN_EXTRARESOURCES": AdalinaModelType.MIN_EXTRARESOURCES,
            "UNSERVED": AdalinaModelType.MINSUM_UNSERVED,
            "UNSERVED_MINMAX": AdalinaModelType.MINMAX_UNSERVED,
            "USAGEFAC_MINMAX": AdalinaModelType.MINMAX_USAGEFAC,
            "MINDIST_ASSIGNMENTS" : AdalinaModelType.MINDIST_ASSIGNMENTS,
            "MINSUM_USAGEFAC" : AdalinaModelType.MINSUM_USAGEFAC,
            "MINSUM_RELOCATION" : AdalinaModelType.MINSUM_RELOCATION,
            "MAXSUM_DIST_UNSERVED" : AdalinaModelType.MAXSUM_DIST_UNSERVED
        }
        if get_str:
            return ', '.join(list(_label_dict.keys()))
        else:
            return _label_dict

    def __str__(self):
        if self == AdalinaModelType.MINSUM_ALLCOST:
            return 'sumcosts'
        elif self == AdalinaModelType.MIN_EXTRARESOURCES:
            return 'misum extra resources'
        elif self == AdalinaModelType.MINSUM_UNSERVED:
            return 'count unserved'
        elif self == AdalinaModelType.MINMAX_UNSERVED:
            return 'minmax unserved'
        elif self == AdalinaModelType.MINMAX_USAGEFAC:
            return 'minmax usage facility'
        elif self == AdalinaModelType.MINDIST_ASSIGNMENTS:
            return 'minsum distance assignments'
        elif self == AdalinaModelType.MINSUM_USAGEFAC:
            return 'minsum usage facility'
        elif self == AdalinaModelType.MINSUM_RELOCATION:
            return 'minsum relocation costs'
        elif self == AdalinaModelType.MAXSUM_DIST_UNSERVED:
            return 'max sum dist unserved'

        return ''

    def get_label(self):

        _label_dict = AdalinaModelType.get_all_labels(False)

        for k, v in _label_dict.items():
            if v == self:
                return k

        return ""

    @staticmethod
    def from_label(lab):

        _label_dict = AdalinaModelType.get_all_labels(False)
        if lab in _label_dict.keys():
            return _label_dict[lab]

        raise ValueError(f"{lab} not recognized as AdalinaModelType!")




from libadalina_analytics.clustering.models.adalina_zoning_data import AdalinaZoningData
from libadalina_analytics.clustering.models.adalina_zoning_solution import AdalinaZoningSolution

import highspy
import numpy as np
import time
import logging
# from sklearn.cluster import AgglomerativeClustering

# from itertools import permutations
# hscb = highspy.cb

# VAR
# x_ij {1 if i in cluster with representative j, 0 otherwise}
# y_j {1 if j is representative of a cluster, 0 otherwise}

# min \sum{(i,j) \in N^2 : i \not= j} c[i,j] * (x_{ij} + w_{ij})
# Kmin <= sum{j in N} y_j <= Kmax
# sum{j in N} x_ij + y_i = 1 forall i in N
# x_ij <= y_i forall i, j in N : i != j
# x_{ij} <= sum{k in N : k in adj(i) and dist^-(k, j) <= dist^-(i,j) } x_{kj} forall (i,j) : i not in adj(j)
# x_{ik} + x_{jk} <= 1 + w_{ij} forall i,j,k in N^3 : i \not= j \not= k
# x_ik + x_jk >= 2*w_ij
# x_{ik} + w_{ik} <= 1 forall (i,k) in N^2 : i \not= k
# \sum_{i in N :  i\not=j } x_ij * c_i >= y_j * F_min forall j in N

class AdalinaZoningModelSimple:

    SIZELIMIT = 150

    def __init__(self, data : AdalinaZoningData,  log_fout = None):

        self.log_fout = log_fout
        self.data = data
        self.vardict = dict()
        self.model = highspy.Highs()
        inf = highspy.kHighsInf
        self.run_status = None
        self.solution = None
        self.best_feasible_sol = None
        self.warmstartsol = None

        # x_ij {1 if i in cluster with representative j, 0 otherwise}
        xvar = []
        for i in self.data.G.nodes:
            for j in self.data.G.nodes:
                if i == j:
                    continue
                xvar.append(
                    self._add_variable(0, 1,
                                       self.data.get_cost_edge(i, j),
                                       highspy.HighsVarType.kInteger,
                                       f"x_{i}_{j}")
                )

        wvar = []
        nnodes = list(self.data.G.nodes)
        for i, el in enumerate(nnodes):
            for j in range(i+1, len(nnodes)):
                el2 = nnodes[j]
                wvar.append(
                    self._add_variable(0, 1,
                                       self.data.get_cost_edge(el, el2),
                                       highspy.HighsVarType.kInteger,
                                       f"w_{i}_{j}")
                )

        # y[s] in {0,1}
        yvar = []
        for i in data.G.nodes:
            yvar.append(
                    self._add_variable(0,1, 0,
                                       highspy.HighsVarType.kInteger,
                                     f"y_{i}")
                                 )

        # K^{\min} <= \sum_{i\in V} z_i <= K^{\max}
        _vars = [self._get_vardict_index(f'y_{i}') for i in data.G.nodes]
        _varvalues = np.ones(self.data.V)
        self.model.addRow(self.data.Kmin, self.data.Kmax,
                          len(_vars),
                          np.array(_vars),
                          np.array(_varvalues)
                          )

        # x_ij <= y_j forall (i,j) in N^2 : i != j
        for i in data.G.nodes:
            for j in data.G.nodes:
                if i == j:
                    continue

                self.model.addRow(-inf,0,
                                  2,
                                  np.array(
                                      [self._get_vardict_index(f'x_{j}_{i}'),
                                       self._get_vardict_index(f'y_{i}')]
                                  ),
                                  np.array([1,-1])
                                  )

        # sum{j in N : i != j} x_ij + y_i = 1 forall i in N
        for i in data.G.nodes:
            _vars = ([self._get_vardict_index(f'x_{i}_{j}') for j in data.G.nodes if i != j] +
                     [self._get_vardict_index(f'y_{i}')])
            self.model.addRow(1, 1,
                              len(_vars),
                              np.array(_vars),
                              np.ones(len(_vars))
                              )

        # x_{ij} <= sum{k in N : k in adj(i) and dist^-(k, j) <= dist^-(i,j) } x_{kj} forall (i,j) : i not in adj(j)
        for i in self.data.G.nodes:
            for j in self.data.G.nodes:
                if i == j or j in self.data.G.neighbors(i):
                    continue

                _vars = [self._get_vardict_index(f'x_{i}_{j}')]
                _varvalues = [1]
                for k in self.data.G.neighbors(i):
                    if self.data.get_node_pair_geographic_distance(i, j) > self.data.get_node_pair_geographic_distance(k, j):
                        _vars.append(self._get_vardict_index(f'x_{k}_{j}'))
                        _varvalues.append(-1)

                self.model.addRow(-inf, 0,
                                  len(_vars),
                                  np.array(_vars),
                                  np.array(_varvalues)
                                  )

        # x_{ik} + x_{jk} <= 1 + w_{ij} forall i,j,k in N^3 : i \not= j \not= k
        # x_{ik} + x_{jk} >= 2*w_{ij} forall i,j,k in N^3 : i \not= j \not= k
        for i, n1 in enumerate(self.data.G.nodes):
            for j, n2 in enumerate(self.data.G.nodes):
                if i >= j:
                    continue

                if i > j:
                    _w = self._get_vardict_index(f'w_{j}_{i}')
                else:
                    _w = self._get_vardict_index(f'w_{i}_{j}')

                for k, n3 in enumerate(self.data.G.nodes):
                    if i == k or j == k:
                        continue

                    self.model.addRow(-inf, 1,
                                      3,
                                      np.array([_w,
                                                self._get_vardict_index(f'x_{n1}_{n3}'),
                                                self._get_vardict_index(f'x_{n2}_{n3}')]),
                                      np.array([-1, 1, 1])
                                      )

                    # self.model.addRow(0, +inf,
                    #                   3,
                    #                   np.array([_w,
                    #                             self._get_vardict_index(f'x_{n1}_{n3}'),
                    #                             self._get_vardict_index(f'x_{n2}_{n3}')]),
                    #                   np.array([-2, 1, 1])
                    #                   )

        # x_{ik} + w_{ik} <= 1 forall (i,k) in N^2 : i \not= k
        for i, n1 in enumerate(self.data.G.nodes):
            for k, n2 in enumerate(self.data.G.nodes):

                if i >= k:
                    continue

                if i > k:
                    _w = self._get_vardict_index(f'w_{k}_{i}')
                else:
                    _w = self._get_vardict_index(f'w_{i}_{k}')

                self.model.addRow(-inf, 1,
                                  3,
                                  np.array([_w,
                                            self._get_vardict_index(f'x_{n1}_{n2}'),
                                            self._get_vardict_index(f'x_{n2}_{n1}')]),
                                  np.array([1, 1, 1])
                                  )

        # \sum_{i in N :  i\not=j } x_ij * c_i >= y_j * F_min forall j in N
        for j in self.data.G.nodes:
            _vars = []
            _varvalues = []
            for i in self.data.G.nodes:
                if i == j:
                    continue
                _vars.append(self._get_vardict_index(f'x_{i}_{j}'))
                _varvalues.append(self.data.get_weight_node(i))

            self.model.addRow(0, inf,
                              len(_vars) + 1,
                              np.array(
                                  [self._get_vardict_index(f'y_{j}')] + _vars
                              ),
                              np.array([-self.data.Fmin] + _varvalues)
                              )


    def _add_variable_dict_element(self, name):
        self.vardict[name] = len(self.vardict)

    def _add_variable(self, lb, ub, obj, vartype, name):
        el = self.model.addVariable(lb, ub,
                                   obj = obj,
                                   type=vartype,
                                   name=name)
        self._add_variable_dict_element(name)

        return el

    def _get_vardict_index(self, name):
        return self.vardict[name]

    def run(self, timelimit=60, label_sol : list = None,
            init_sol : AdalinaZoningSolution = None):

        self.solution = None
        self.model.setOptionValue("time_limit",
                                  timelimit)
        self.warmstartsol = init_sol

        _constr_sol = None
        if label_sol is None and init_sol is None:
            heur_start = time.time()
            logging.debug("start constructive heuristic")
            _constr_sol = AdalinaZoningSolution(self.data)
            if _constr_sol.constructive():
                label_sol = _constr_sol.labels_
                self.warmstartsol = _constr_sol
            else:
                logging.debug("constructive heuristic FAILED")
            logging.debug(f"constructive heuristic ended in {time.time() - heur_start}")

        elif label_sol is None and init_sol is not None:
            label_sol = init_sol.labels_

        if label_sol is not None:
            xwarmstart = []
            for el in np.unique(label_sol):
                areas = np.where(label_sol==el)[0]
                xwarmstart.append(self._get_vardict_index(f'y_{areas[0]}'))
                for i, _el in enumerate(areas):
                    for j in range(i+1, areas.shape[0]):
                        xwarmstart.append(self._get_vardict_index(f'x_{_el}_{areas[j]}'))

            xwarmstart = np.array(xwarmstart, dtype=int)
            self.model.setSolution(xwarmstart.shape[0], xwarmstart, np.ones(xwarmstart.shape[0]))

        # start_time = time.time()
        if self.data.V > self.SIZELIMIT:
            if _constr_sol is not None:
                self.solution  = _constr_sol
            print(f"WARNING: istanza troppo grande, soluzione solo euristica {self.data.V} > {self.SIZELIMIT}")
            return None

        self.solution = None
        self.run_status = self.model.run()

        self.get_solution(run_model=False)

        return self.run_status

    def _build_solution(self):
        self.solution = AdalinaZoningSolution(self.data)
        allvars = self.model.getVariables()

        _repr = []
        for varname, varindex in self.vardict.items():

            if not varname.startswith("y_"):
                continue

            val = self.model.variableValue(allvars[varindex])
            if val < 1e-5:
                continue

            l = varname.split("_")

            _repr.append(int(l[1]))

        for r in _repr:
            self.solution.add_repr(r, True)

        for varname, varindex in self.vardict.items():

            if not varname.startswith("x_"):
                continue

            val = self.model.variableValue(allvars[varindex])
            if val < 1e-5:
                continue

            l = varname.split("_")

            self.solution.add_belonging(int(l[1]), int(l[2]))

        for varname, varindex in self.vardict.items():

            if not varname.startswith("w_"):
                continue

            val = self.model.variableValue(allvars[varindex])
            if val < 1e-5:
                continue

            l = varname.split("_")

            if not self.solution.check_nodes_belong_same_cluster([int(l[1]), int(l[2])]):
                continue

            self.solution.add_edge(int(l[1]), int(l[2]))

        return self.solution.check_feasibility()

    def get_solution(self, run_model=True):

        if self.solution is not None:
            return self.solution

        info = self.model.getInfo()
        if (self.run_status == highspy.HighsStatus.kOk or
              (self.run_status == highspy.HighsStatus.kWarning and
               info.primal_solution_status == highspy.SolutionStatus.kSolutionStatusFeasible)):

            logging.debug(f"obj. fun. value {self.model.getObjectiveValue()}")

            if self._build_solution():
                logging.debug("solution feasible!")
                return self.solution
            else:
                logging.debug("model solved but solution infeasible!")

        elif self.run_status == highspy.HighsStatus.kError:

            logging.debug("model ended with Error")

        if self.solution is None and run_model:
            self.run()
            return self.get_solution(run_model=False)

        if self.solution is not None:
            return self.solution
        elif self.warmstartsol is not None:
            return self.warmstartsol
        else:
            logging.debug("no solution available. run model with function run()")
            return None



from .adalina_data import AdalinaData
from .adalina_model_type import AdalinaModelType
from .adalina_solution import AdalinaSolution
from .options import AdalinaAlgorithmOptions
from .relocation_resource import RelocationResource

__all__ = [
    'AdalinaData',
    'AdalinaModelType',
    'AdalinaSolution',
    'AdalinaAlgorithmOptions',
    'RelocationResource'
]



from libadalina_analytics.utils import GeometryFormats
from .zoning_model_simple import AdalinaZoningModelSimple
from libadalina_core.sedona_utils import DataFrame, EPSGFormats

from ..models.adalina_zoning_data import AdalinaZoningData
from ..models.adalina_zoning_distance import ClusteringDistance
from ..models.adalina_zoning_solution import AdalinaZoningSolution


def clustering_algorithm(data: DataFrame,
                         epsg: EPSGFormats,
                         geometry_column: str = 'geometry',
                         geometry_format: GeometryFormats = GeometryFormats.WKT,
                         weight_column: str | None = None,
                         k_min: int | None = None,
                         k_max: int | None = None,
                         f_min: float | None = None,
                         distances: list[ClusteringDistance] | None = None,
                         timelimit: int = 60) -> AdalinaZoningSolution | None:
    """
    Create clusters of similar areas minimizing their internal distance.

    Distances are given as a list of ClusteringDistance objects, each containing:
    - name: the name of the column in the input data
    - weight: the weight of the distance in the overall distance calculation
    - function: the distance function to use

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame or pyspark.sql.DataFrame
        The input data containing the geometries and attributes to be clustered.
    epsg : EPSGFormats
        The EPSG format of the input geometries.
    geometry_column : str
        The name of the column containing the geometries. Default is 'geometry'.
    geometry_format : GeometryFormats
        The format of the geometries in the geometry column. Default is GeometryFormats.WKT.
    weight_column : str | None
        The name of the column to use as the optional weight of an area. Default is None.
    k_min : int | None
        The minimum number of clusters to create. If None, defaults to 1.
    k_max : int | None
        The maximum number of clusters to create. Default is None.
    f_min : float | None
        The minimum total weight of a cluster. If None, defaults to 1.
    distances : list[ClusteringDistance] | None
        List of distances to use for clustering. If None, no only distance between area centroids will be used.
        Each distance is represented as a ClusteringDistance object that includes the name of the column,
        the weight of the distance, and the distance function to use.
    timelimit : int
        The maximum time (in seconds) to run the algorithm. Default is 60 seconds.

    Returns
    -------
    AdalinaZoningSolution

    """
    user_input: dict = {
        "epsg": epsg,
        "geometry_type": geometry_format,
        "geometry": geometry_column,
        "timelimit": timelimit
    }
    if weight_column is not None:
        user_input["weight"] = weight_column
    if k_min is not None:
        user_input["Kmin"] = k_min
    if k_max is not None:
        user_input["Kmax"] = k_max
    if f_min is not None:
        user_input["Fmin"] = f_min
    if distances is not None:
        user_input["distances"] = [{
            'name': distance.name,
            'weight': distance.weight,
            'func': distance.function
        } for distance in distances]

    data = AdalinaZoningData.from_Amelia(amelia_file=data, user_input=user_input)

    model = AdalinaZoningModelSimple(data)

    _ = model.run(timelimit=timelimit)

    sol = model.get_solution(False)

    return sol




import networkx as nx
import geopandas as gpd
import pandas as pd
import numpy as np
from pandas.api.types import is_float_dtype, is_integer_dtype, is_string_dtype
from sklearn.metrics import DistanceMetric
import time
# from collections import defaultdict
import logging
import shapely

from networkx.algorithms import flow
from scipy.spatial import distance_matrix, distance
# import gower
import json, jsonschema

from libadalina_core.sedona_utils import EPSGFormats
from libadalina_analytics.utils import GeometryFormats

EPSG_FORMAT_FOR_ALGORITHM = EPSGFormats.EPSG32632

class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


def validate_json(input_dict : dict, schema_path : str, log_fout = None):
    with open(schema_path, "r") as fin:
        input_schema = json.load(fin)
    try:
        jsonschema.validate(instance=input_dict, schema=input_schema)  # Validazione
        return True
    except jsonschema.exceptions.ValidationError as e:
        logging.error("Errore di validazione input utente: %s " % e.message, log_fout)
        return False

def validate_user_input_json(user_input : dict, log_fout = None):
    return validate_json(user_input, "adalina_zoning_input_schema.json", log_fout=log_fout)

def validate_solution_json(sol_dict : dict, log_fout = None):
    return validate_json(sol_dict, "adalina_zoning_solution_schema.json", log_fout=log_fout)

from shapely.geometry import Polygon

def get_all_vertices(poly : Polygon):
    coords = list(poly.exterior.coords)
    for interior in poly.interiors:
        coords.extend(interior.coords)
    return coords

def extract_vertices(geom):
    if geom.geom_type == 'Polygon':
        return get_all_vertices(geom)
    elif geom.geom_type == 'MultiPolygon':
        all_coords = []
        for poly in geom.geoms:
            all_coords.extend(get_all_vertices(poly))
        return all_coords

    return []

def polygons_vertices_distance(polys : [Polygon]):

    n = len(polys)
    d = np.zeros((n,n))

    polyvert = [get_all_vertices(p) for p in polys]

    for i in range(n-1):
        for j in range(i+1, n):
            d[i,j] = distance.cdist(polyvert[i], polyvert[j]).min()

    return d


class AdalinaZoningData:

    AMELIA_COLUMNS_MANDATORY = [
        "geometry",
    ]

    DISTANCE_FUNCTION_PER_TYPES = {
        "numeric" : ["euclidean", "manhattan", "chebyshev"],
        "geometry" : ["haversine"],
        "string" : ["hamming", "canberra", "braycurtis"],
        "bool" : ["jaccard", "matching", "dice", "kulsinski", "rogerstanimoto", "russellrao", "sokalmichener", "sokalsneath"]
    }

    def __init__(self,
                 geodataFrame : gpd.GeoDataFrame = None,
                 G : nx.Graph = None,
                 nodes_costs : pd.DataFrame = None,
                 user_input : dict = None,
                 log_fout = None):

        if user_input is None:
            user_input = dict()

        self.log_fout = log_fout
        self.gdf = geodataFrame

        self.G = G
        self.node_pairs_costs_df = None
        self.min_geodistance_polygons = None
        self._from_index_to_int = None
        self.old_index_name = None

        if self.gdf is not None:

            self.old_index_name = self.gdf.index.name
            self.gdf.reset_index(inplace=True, names="old_index")

            geom_vect = self.gdf.geometry.array
            n = len(geom_vect)
            self._from_index_to_int = dict()
            for i, idx in enumerate(self.gdf.index):
                self._from_index_to_int[idx] = i

            logging.debug("start computing distance between polygons")
            _st = time.time()

            # prendo i centroidi delle geometrie degli ospedali
            _centroids = self.gdf.centroid
            # matrice NxN di distanze tra tutte le coppie di ospedali (in chilometri)
            self.min_geodistance_polygons = _centroids.apply(lambda g: _centroids.distance(g)).to_numpy() / 1000

            #self.min_geodistance_polygons = np.zeros((n, n))
            #for i in range(n - 1):
            #    logging.debug(f"distance polygon index {i + 1} over {n}")
            #    self.min_geodistance_polygons[i, (i + 1):] = geom_vect[i].distance(geom_vect[(i + 1):])
            logging.debug(f"distance computed in {time.time() - _st}")

            G = nx.Graph()

            for idx in self.gdf.index:
                G.add_node(idx)

            sindex = self.gdf.sindex

            for i, geom in enumerate(self.gdf.geometry):
                possible_matches_idx = list(sindex.intersection(geom.bounds))
                possible_matches = self.gdf.loc[possible_matches_idx]
                possible_matches = possible_matches.loc[possible_matches.geometry.intersects(geom)]
                for j in possible_matches.index:
                    if i < j: # and geom.touches(geodataFrame.geometry[j]):
                        G.add_edge(i, int(j))

            if "weight" in user_input:
                for idx, row in self.gdf.iterrows():
                    G.nodes[idx]["weight"] = row["weight"]

            self.G = G

            if "distances" not in user_input:
                # Estrai coordinate dei centroidi
                centroid_coords = self.gdf.centroid.apply(lambda p: (p.x, p.y)).tolist()
                # Crea la matrice con SciPy
                self.node_pairs_costs_df = distance_matrix(centroid_coords, centroid_coords)
                # sum_weight_gower = 1
            else:
                sum_weight_gower = 0
                all_dist = []
                for el in user_input["distances"]:
                    col = el["name"]
                    weight = el.get("weight", 1)
                    func = el.get("func", None)

                    val = None

                    if func is not None:
                        if isinstance(self.gdf.dtypes[col], gpd.array.GeometryDtype) and func in AdalinaZoningData.DISTANCE_FUNCTION_PER_TYPES["geometry"]:
                            centroid_coords = self.gdf.centroid.apply(lambda p: (p.x, p.y)).tolist()
                            val = distance_matrix(centroid_coords, centroid_coords)

                        elif ((is_float_dtype(self.gdf.dtypes[col]) or is_integer_dtype(self.gdf.dtypes[col])) and func in AdalinaZoningData.DISTANCE_FUNCTION_PER_TYPES["numeric"]) or \
                            (is_string_dtype(self.gdf.dtypes[col]) and func in AdalinaZoningData.DISTANCE_FUNCTION_PER_TYPES["string"]) or \
                            (self.gdf.dtypes[col] == bool or self.gdf.dtypes[col] == int and func in AdalinaZoningData.DISTANCE_FUNCTION_PER_TYPES["bool"]):

                            val = DistanceMetric.get_metric(func).pairwise(self.gdf[[col]])
                        else:
                            func = None

                    if func is None:
                        if isinstance(self.gdf.dtypes[col], gpd.array.GeometryDtype):
                            centroid_coords = self.gdf.centroid.apply(lambda p: (p.x, p.y)).tolist()
                            val = distance_matrix(centroid_coords, centroid_coords)
                        elif is_float_dtype(self.gdf.dtypes[col]) or is_integer_dtype(self.gdf.dtypes[col]):
                            val = DistanceMetric.get_metric("euclidean").pairwise(self.gdf[[col]])
                        elif is_string_dtype(self.gdf.dtypes[col]):
                            cat_series = pd.Categorical(self.gdf[col])
                            val = DistanceMetric.get_metric("hamming").pairwise(pd.DataFrame(cat_series.codes))
                        elif self.gdf.dtypes[col] == bool or self.gdf.dtypes[col] == int:
                            val = DistanceMetric.get_metric("jaccard").pairwise(self.gdf[[col]].astype(bool))

                    if val is not None:
                        val = (val - val.min()) / (val.max() - val.min())
                        all_dist.append(
                            val * weight
                        )

                        sum_weight_gower += weight

                all_dist = np.array(all_dist)
                all_dist = (all_dist - all_dist.min()) / (all_dist.max() - all_dist.min())

                self.node_pairs_costs_df = all_dist.sum(axis=0)/sum_weight_gower

            self.node_pairs_costs_df = pd.DataFrame(self.node_pairs_costs_df,
                                                    index=self.gdf.index,
                                                    columns=self.gdf.index)

        elif G is not None and nodes_costs is not None:
            self.G = G

            self.gdf = pd.DataFrame.from_dict({'old_index' : nodes_costs.index.values})

            self.node_pairs_costs_df = nodes_costs

            self.node_pairs_costs_df.reset_index(inplace=True, drop=True)

            mp = {}
            for i, row in self.gdf.iterrows():
                mp[row.old_index] = i
            self.G = nx.relabel_nodes(G, mp)

            self.min_geodistance_polygons = nodes_costs.to_numpy()

            self._from_index_to_int = dict()
            for i, idx in enumerate(self.gdf.index):
                self._from_index_to_int[idx] = i

        else:
            raise ValueError("either 'geodataFrame' or both 'G' and 'nodes_costs' must have value")

        self.E1 = []
        self.edges = dict()
        for i in self.G.nodes:
            self.edges[i] = []
            for j in self.G.nodes:
                if j <= i:
                    continue

                self.E1.append((i, j))
                self.edges[i].append(j)

        self.T = None

        self.Fmin = user_input.get("Fmin", 1)
        self.Kmin = user_input.get("Kmin", 1)
        self.V = len(self.G.nodes)
        self.Kmax = user_input.get("Kmax", self.V)
        self.Kmax = min(self.Kmax, self.V)

        if self.Kmin > self.Kmax:
            self.Kmin = self.Kmax

    @classmethod
    def from_shapefile(cls, filename : str):

        gdf = gpd.read_file(filename)
        gdf.reset_index(inplace=True, drop=True)

        return cls(gdf)

    @classmethod
    def from_nxGraph(cls, graph: nx.Graph,
                     nodes_costs : pd.DataFrame):

        return cls(G = graph, nodes_costs=nodes_costs)

    @classmethod
    def from_Amelia(cls, amelia_file : pd.DataFrame,
                    user_input : dict, log_fout = None):

        # if not validate_user_input_json(user_input, log_fout):
        #     raise ValueError("Error parsing user input, check log for more information")

        missing_columns = [el for el in AdalinaZoningData.AMELIA_COLUMNS_MANDATORY
                           if el not in user_input.keys()]

        if len(missing_columns) > 0:
            raise ValueError("Error parsing from Amelia dataset: missing columns: " + ' '.join(missing_columns))

        geometry_type = GeometryFormats.WKT
        if 'geometry_type' in user_input:
            geometry_type = user_input["geometry_type"]

        epsg = EPSGFormats.EPSG4326
        if "epsg" in user_input:
            epsg = user_input["epsg"]

        col_rename = {
            user_input["geometry"]: "geometry"
        }
        if "weight" in user_input:
            col_rename[user_input["weight"]] = "weight"

        amelia_file.rename(columns=col_rename, inplace=True)

        amelia_file.reset_index(inplace=True, drop=True)
        amelia_file = AdalinaZoningData.from_dataframe_to_geopandas(amelia_file, epsg, "geometry", geometry_type)

        # debug!
        # amelia_file = amelia_file.iloc[7:]

        return cls(amelia_file, user_input=user_input)

    def get_node_pair_geographic_distance(self, node1, node2):
        if node1 == node2:
            return 0

        i = self._from_index_to_int[node1]
        j = self._from_index_to_int[node2]

        if i > j:
            i, j = j, i

        return self.get_pair_geographic_distance(i, j)

    def get_pair_geographic_distance(self, i, j):

        if i == j:
            return 0

        return self.min_geodistance_polygons[i,j]

    def get_cost_edge(self, i, j):

        return self.node_pairs_costs_df.loc[i,j]

    def get_weight_node(self, i):

        if "weight" not in self.G.nodes[i]:
            return 1

        return self.G.nodes[i]["weight"]

    def gomory_hu(self, weight : dict):

        for e in self.G.edges:
            w = weight.get(e, 0)
            self.G.edges[e]["weight"] = w

        # for k, v in weight.items():
        #     self.G.edges[k]["weight"] = v

        self.T = nx.gomory_hu_tree(self.G, capacity = "weight", flow_func=flow.boykov_kolmogorov)

        return self.T

    def minimum_edge_weight_in_shortest_path(self, u, v):

        return nx.minimum_cut(self.T, u, v, capacity ="weight")

        # path = nx.shortest_path(self.T, u, v, weight="weight")

        # return min((self.T[u][v]["weight"], (u, v)) for (u, v) in zip(path, path[1:]))

    def check_nodes_subset_connected(self, subset_nodes):
        if len(subset_nodes) < 2:
            return True

        subG = self.G.subgraph(subset_nodes)
        return nx.is_connected(subG)

    @staticmethod
    def from_dataframe_to_geopandas(df : pd.DataFrame,
                                    epsg : EPSGFormats = EPSGFormats.EPSG4326,
                                    wkt_col : str = "geometry",
                                    geometry_format: GeometryFormats = GeometryFormats.WKT):
        gdf = gpd.GeoDataFrame(df)

        if geometry_format == GeometryFormats.WKT:
            gdf["geometry"] = gpd.GeoSeries.from_wkt(df[wkt_col])
        elif geometry_format == GeometryFormats.GEOJSON:
            gdf["geometry"] = df[wkt_col].apply(shapely.from_geojson)

        gdf = gdf.set_geometry(col="geometry")
        gdf = gdf.set_crs(epsg=epsg.value)
        if epsg != EPSG_FORMAT_FOR_ALGORITHM:
            gdf = gdf.to_crs(epsg=EPSG_FORMAT_FOR_ALGORITHM.value)

        return gdf

    def add_clustering(self, clustering : dict):

        # cl_dict = {}
        # kindex = 0
        # for k, v in clustering.items():
        #     for el in v:
        #         cl_dict[el] = kindex
        #     kindex += 1

        res = pd.Series(clustering) # cl_dict)
        ret_gdf = None
        if self.gdf is not None:
            ret_gdf = self.gdf.set_index("old_index")
            ret_gdf.index.name = self.old_index_name
            ret_gdf["zoning_Adalina"] = res


        return ret_gdf




import logging

from ..models import AdalinaData, AdalinaModelType, AdalinaAlgorithmOptions, AdalinaSolution
from .adalina_model_highs import AdalinaModelHighs
import time

def run_model_asis(problem, options):

    assert isinstance(problem, AdalinaModelHighs)

    logging.debug(f"RUN MODEL type {problem.modeltype.get_label()}")

    if problem.run(timelimit=options.timelimit):
        logging.debug(f"solved with o.f. value {problem.objective_value}!")  # problem.model.getObjectiveValue()}

        # if options.outdir is not None:
        #
        #     if solfile_suffix != "" and not solfile_suffix.startswith("_"):
        #         solfile_suffix = "_" + solfile_suffix
        #
        #     problem.get_solution().export_json(options.outdir +
        #                                        options.fileprefix +
        #                                        "_" + problem.modeltype.get_label() + solfile_suffix + '_sol')

        return problem.get_solution()

    else:
        logging.warning("SOLUTION NOT AVAILABLE")

    return None

def run_hierarchy_with_distance_threshold(data : AdalinaData,
                                          options : AdalinaAlgorithmOptions,
                                          run_it : int = 1,
                                          threshold_distance_relocation : float = None,
                                          problem : AdalinaModelHighs = None):
    """

    :param data: AdalinaData object
    :param options: AdalinaAlgorithmOptions object
    :param run_it: integer, default to 1, for iterations of runs
    :param threshold_distance_relocation: float, default to None, override threshold of maximum relocation
    :param problem: AdalinaModelHighs object, default to None, for iterations of run
    :return: AdalinaModelHighs object (or None if no solution available), [AdalinaSolution] list of solutions available
    """

    all_solutions = []

    start_step1 = time.perf_counter()
    logging.debug(f"STEP {run_it}.1: MIN SUM UNSERVED DEMAND")

    options.change_model_type(AdalinaModelType.MINSUM_UNSERVED)
    if problem is None:
        problem = AdalinaModelHighs(data, options.modeltype)
    else:
        problem.reset_all_objcoeff()
        problem.reset_solution()
        problem.set_model_minsum_unserved()

    if threshold_distance_relocation is not None:
        if threshold_distance_relocation < 1e-5:
            # forbid relocation
            problem.set_rhovar_to_zero()
        elif threshold_distance_relocation > 1e-5:
            problem.set_rhovar_by_distance_threshold(threshold_distance_relocation)

    """

    1 - min sum domanda non servita
    2 - min sum distanza di assegnamento (con UB su 1)
        domanda non servita per zona è KPI di rischio della zona per l'evento catastrofico

    """

    sol_minsum_unserved = run_model_asis(problem, options)# , f"_step{run_it}")
    all_solutions.append(sol_minsum_unserved)

    end_step1 = time.perf_counter()
    logging.debug(f"END STEP {run_it}.1 IN {(end_step1-start_step1):.3f} secondi")

    if sol_minsum_unserved is None:
        logging.debug("ERR: NO SOLUTION FOR min sum unserved demand")
        return None, all_solutions

    assert isinstance(sol_minsum_unserved, AdalinaSolution)

    problem.data.ub_total_unserved_demand = sol_minsum_unserved.objfun_value

    # remove obj. fun. coefficient on variables w
    problem.reset_wvar_costs()
    # UB on objective 1
    problem.set_constr_ub_sumunserved(problem.data.ub_total_unserved_demand)

    if options.run_parametric_analysis:
        if sol_minsum_unserved.objfun_value < 1e-5:
            logging.debug("NO UNSERVED DEMAND END HIERARCHY")

            if run_it <= 1:
                return problem, all_solutions

        else:

            start_step2a = time.perf_counter()
            problem.reset_solution()
            # 2 - min sum assignment distance (with UB on objective 1)
            logging.debug(f"STEP {run_it}.2: MIN SUM UNASSIGNMENT DISTANCE")
            problem.modify_model_to_minsumdist()
            options.change_model_type(AdalinaModelType.MINDIST_ASSIGNMENTS)

            # problem.modify_model_to_maxdist_unserved()
            # options.change_model_type(AdalinaModelType.MAXSUM_DIST_UNSERVED)

            # problem.model.writeModel("/home/marco/Desktop/test_maxsumdistuns.lp")
            sol_min_sumdist_assign = run_model_asis(problem, options)# ,f"_step{run_it}")

            end_step2a = time.perf_counter()
            logging.debug(f"END STEP {run_it}.2 IN {(end_step2a - start_step2a):.3f} secondi")

            if sol_min_sumdist_assign is None:
                logging.debug("ERR: NO SOLUTION FOR min sum assignment distance")
                return problem, all_solutions

            assert isinstance(sol_min_sumdist_assign, AdalinaSolution)

            all_solutions.append(sol_min_sumdist_assign)
            # export_html(options, sol_min_sumdist_assign, data, filesuffix=f"step{run_it}")

            problem.reset_x_val_cost()

    if run_it <= 1 and options.run_parametric_analysis:

        return problem, all_solutions

    """

    B - permetti relocation
      - analisi parametrica sulla distanza soglia che permette relocation. per ogni soglia:
            1 - min sum domanda non servita
            2a - min sum distanza di assegnamento (con UB su 1)
                - come riassumere i risultati di tutta l'analisi parametrica in valori di KPI?
                    * distanza per arrivare a zero unserved
                    * distanza per soglie valueable di unserved percentuali?
            2b - min max utilizzo relativo facility (con UB su 1)
            3b - min sum distanza di assegnamento (con UB su 1 e 2b)
            4b - min sum distanza di relocation (con UB su 1, 2b, 3b e 4b)

    """

    """
    HIERARCHY STEP 2b - min max utilizzo relativo facility (con UB su 1)
    """
    # reset xvar val
    start_step2b = time.perf_counter()

    problem.reset_solution()
    logging.debug(f"STEP {run_it}.2b MIN MAX FACILITY USAGE")
    problem.modify_model_to_minmax_usage_shelter()
    options.change_model_type(AdalinaModelType.MINMAX_USAGEFAC)

    # problem.model.writeModel("/home/marco/Desktop/test_minmaxrho.lp")
    sol_minmaxusage = run_model_asis(problem, options)# , f"_step{run_it}")

    end_step2b = time.perf_counter()
    logging.debug(f"END STEP {run_it}.2b IN {(end_step2b - start_step2b):.3f} secondi")

    if sol_minmaxusage is None:
        logging.debug("ERR: NO SOLUTION FOR min max usage of facility")
        return problem, all_solutions

    all_solutions.append(sol_minmaxusage)

    problem.reset_solution()
    problem.reset_rhomax_var_objfun_coeff()
    problem.set_constr_ub_maxusage(sol_minmaxusage.objfun_value)

    """
    HIERARCHY STEP 3 MIN SUM ASSIGNMENT DISTANCE
    """
    start_step3 = time.perf_counter()

    logging.debug(f"STEP {run_it}.3 MIN SUM ASSIGNMENT DISTANCE")
    problem.modify_model_to_minsumdist()
    options.change_model_type(AdalinaModelType.MINDIST_ASSIGNMENTS)

    # problem.model.writeModel("/home/marco/Desktop/test_misumdist.lp")
    sol_minsumdist = run_model_asis(problem, options) #, f"_step{run_it}")

    end_step3 = time.perf_counter()
    logging.debug(f"END STEP {run_it}.3 IN {(end_step3 - start_step3):.3f} secondi")

    if sol_minsumdist is None:
        logging.debug("ERR: NO SOLUTION FOR min sum assignment distance")
        return problem, all_solutions

    all_solutions.append(sol_minsumdist)

    assert isinstance(sol_minsumdist, AdalinaSolution)

    problem.data.ub_total_assignment_distance = sol_minsumdist.objfun_value
    problem.reset_solution()
    problem.reset_x_val_cost()
    problem.set_constr_ub_sumdist()

    """
    HIERARCHY STEP 4 MIN SUM RELOCATION
    """
    start_step4 = time.perf_counter()

    logging.debug(f"STEP {run_it}.4 MIN SUM RELOCATION COST")
    problem.modify_model_to_minsum_relocation_costs()
    options.change_model_type(AdalinaModelType.MINSUM_RELOCATION)

    # problem.model.writeModel("/home/marco/Desktop/test_minsumrho.lp")
    sol_minsumrho = run_model_asis(problem, options)#, f"_step{run_it}")

    end_step4 = time.perf_counter()
    logging.debug(f"END STEP {run_it}.4 IN {(end_step4 - start_step4):.3f} secondi")

    if sol_minsumrho is None:
        logging.debug("ERR: NO SOLUTION FOR min sum relocation costs")
        return problem, all_solutions

    all_solutions.append(sol_minsumrho)
    # export_html(options, sol_minsumrho, data, filesuffix=f"step{run_it}")

    return problem, all_solutions




import time

class Timing:

    def __init__(self, message: str = 'Running time {.1:f}'):
        self.__message = message
        self.__start = None

    def __enter__(self):
        self.__start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed = time.time() - self.__start
        print(self.__message.format(elapsed))


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
import pathlib

sys.path.insert(0, os.path.abspath(pathlib.Path(__file__).parent.parent.parent))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'libadalina-analytics'
copyright = '2025, OptLab, University of Milan'
author = 'Marco Casazza, Alberto Ceselli, Marco Premoli'
release = '1.1.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    "sphinxcontrib.collections",
    'nbsphinx',
    'myst_parser'
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Intersphinx settings
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable', None),
    'geopandas': ('https://geopandas.org/en/stable/', None),
    'pyspark': ('https://spark.apache.org/docs/latest/api/python/', None),
}

collections = {
    # 'notebooks': {
    #     'driver': 'copy_folder',
    #     'source': pathlib.Path(__file__).parent / 'libadalina-examples',
    #     'target': 'notebooks',
    #     'ignore': ['*.py', '.sh'],
    # }
}

templates_path = ['_templates']
exclude_patterns = ['_build', '**.ipynb_checkpoints']

nbsphinx_execute = 'never'  # Do not execute notebooks during build
highlight_language = 'python'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'pydata_sphinx_theme'
html_static_path = ['_static']

latex_elements = {
    'papersize': 'a4paper',
    'pointsize': '10pt',
    'extraclassoptions': 'oneside',
    'fncychap': r'\usepackage[Bjornstrup]{fncychap}',
    'makeindex': r'\usepackage[columns=1]{idxlayout}\makeindex'
}


import itertools
import math

import shapely
from libadalina_core.sedona_configuration import get_sedona_context
from libadalina_core.sedona_utils import DataFrame, to_spark_dataframe, DEFAULT_EPSG

from libadalina_analytics.flows_distribution.algorithms.origin_destination_extractor import get_shape_node
import networkx as nx
import dataclasses
import pandas as pd
import geopandas as gpd
import pyspark.sql as ps 
from pyspark.sql import functions as func
import gc

@dataclasses.dataclass
class GraphCost:
    name: str
    """The name of the cost attribute in the graph edges."""
    cost_per_unit: float
    """The cost per unit of the cost attribute."""
    weight: float
    """The weight of the cost attribute in the overall cost calculation. The sum of all weights must be equal to 1.0."""


def _compute_edge_cost(graph: nx.Graph, edge, edge_data, graph_costs: list[GraphCost]) -> float:
    cost = sum(
        edge_data.get(cost.name, 0) * cost.cost_per_unit * cost.weight for cost in graph_costs
    )
    graph.edges[edge]['__cost'] = cost
    return cost


def _compute_edges_cost(graph: nx.Graph, graph_costs: list[GraphCost]) -> None:
    for edge in graph.edges(data=True):
        _compute_edge_cost(graph, edge[0:2], edge[2], graph_costs)


def _get_source_and_destination_pairs(
        graph: nx.Graph,
        shapes_df: DataFrame,
        flows_df: DataFrame,
        shapes_id_column: str = 'id',
        flows_origin_id_column: str = 'origin_id',
        flows_destination_id_column: str = 'destination_id',
        sources: list = None,
        destinations: list = None
) -> ps.DataFrame:

    # filter flows DataFrame if sources or destinations are provided
    if sources is not None:
        flows_df = flows_df.filter(func.col(flows_origin_id_column).isin(sources))
    if destinations is not None:
        flows_df = flows_df.filter(func.col(flows_destination_id_column).isin(destinations))

    # filter out the flows from the same origin to the same destination
    flows_df = (flows_df
                .filter(func.col(flows_origin_id_column) != func.col(flows_destination_id_column)))

    shapes_df = to_spark_dataframe(shapes_df)
    if sources is not None or destinations is not None:
        sources_ids = (row[0] for row in flows_df.select(func.col(flows_origin_id_column)).distinct().collect())
        destination_ids = (row[0] for row in flows_df.select(func.col(flows_destination_id_column)).distinct().collect())

        shapes_df = shapes_df.filter(func.col(shapes_id_column).isin(list(set(
            itertools.chain(sources_ids, destination_ids)
        ))))
    shapes_node = get_shape_node(graph, shapes_df, shapes_id_column, [])

    # join with the tables to get the origin and destination graph nodes
    return (flows_df
                .join(shapes_node, func.col(flows_origin_id_column) == func.col(shapes_id_column), how='inner')
                .select(*(list(flows_df.columns)), func.col('closest_node').alias('origin_node'))
                .join(shapes_node, func.col(flows_destination_id_column) == func.col(shapes_id_column), how='inner')
                .select(*(list(flows_df.columns)), func.col('origin_node'),
                        func.col('closest_node').alias('destination_node'))
                )

def _compute_shortest_path(graph: nx.Graph, graph_costs: list[GraphCost], origin_id, destination_id, origin_node, destination_node):
    try:
        path = nx.astar_path(
            graph,
            source=origin_node,
            target=destination_node,
            weight='__cost'
        )
        unitary_cost = nx.path_weight(graph, path, weight='__cost')
        geometry = shapely.MultiLineString(
            [graph.get_edge_data(path[i], path[i + 1])['geometry'] for i in range(len(path) - 1)])
        return [origin_id, destination_id, path, geometry, unitary_cost] + [nx.path_weight(graph, path, weight=cost.name) for cost in graph_costs]
    except nx.NetworkXNoPath:
        return [origin_id, destination_id, [], shapely.Point(), math.inf] + [math.inf for _ in graph_costs]
    finally:
        gc.collect()

def flows_distribution_algorithm(
        graph: nx.Graph,
        shapes_df: DataFrame,
        flows_df: DataFrame,
        graph_costs: list[GraphCost],
        shapes_id_column: str = 'id',
        flows_origin_id_column: str = 'origin_id',
        flows_destination_id_column: str = 'destination_id',
        flows_demand_column: str = 'demand',
        sources: list = None,
        destinations: list = None
) -> ps.DataFrame | gpd.GeoDataFrame:
    """
    Distribute flow demands between origin and destination areas on a graph while minimizing total transportation costs.

    Parameters
    ----------
    graph : nx.Graph
        The networkx graph on which to distribute the flows. The graph edges must contain the cost attributes
        specified in `graph_costs`.
    shapes_df : pandas.DataFrame or geopandas.GeoDataFrame or pyspark.DataFrame
        A DataFrame containing the shapes of the areas with their geometries and identifiers.
    flows_df : pandas.DataFrame or geopandas.GeoDataFrame or pyspark.DataFrame
        A DataFrame containing the flows to distribute, with origin and destination identifiers and demand values.
    graph_costs : list[GraphCost]
        A list of GraphCost objects defining the cost attributes to consider in the graph edges, their cost per unit,
        and their weights in the overall cost calculation. The sum of all weights must be equal to 1.0.
    shapes_id_column : str
        The name of the column in `shapes_df` containing the area identifiers. Default is 'id'.
    flows_origin_id_column : str
        The name of the column in `flows_df` containing the origin area identifiers. Default is 'origin_id'.
    flows_destination_id_column : str
        The name of the column in `flows_df` containing the destination area identifiers. Default is 'destination_id'.
    flows_demand_column : str
        The name of the column in `flows_df` containing the demand values to distribute. Default is 'demand'.
    sources : list | None
        An optional list of origin area identifiers to filter the flows. If provided, only flows with origins in this
        list will be considered. Default is None.
        Identifiers must be of the same type as the ones in `flows_origin_id_column`.
    destinations : list | None
        An optional list of destination area identifiers to filter the flows. If provided, only flows with destinations
        in this list will be considered. Default is None.
        Identifiers must be of the same type as the ones in `flows_destination_id_column`.

    Returns
    -------
    pandas.DataFrame or geopandas.GeoDataFrame or pyspark.DataFrame
        The DataFrame flows_df enriched with columns for the computed path, its geometry,
        the total path cost, and the individual costs for each pair or origin and destination areas.
        If the input `flows_df` is a pandas DataFrame or a geopandas GeoDataFrame, the output will be a geopandas
        GeoDataFrame, otherwise a pyspark DataFrame.
    """
    if sum(cost.weight for cost in graph_costs) != 1.0:
        raise ValueError("The sum of the weights of the graph costs must be equal to 1.0")

    flows_df_type = type(flows_df)

    if isinstance(flows_df, pd.DataFrame) or isinstance(flows_df, gpd.GeoDataFrame):
        flows_df = get_sedona_context().createDataFrame(flows_df)

    source_and_destination_pairs = _get_source_and_destination_pairs(
        graph, shapes_df, flows_df,
        shapes_id_column, flows_origin_id_column, flows_destination_id_column,
        sources, destinations
    ).select(flows_origin_id_column, flows_destination_id_column, 'origin_node',
                                        'destination_node').distinct()

    _compute_edges_cost(graph, graph_costs)

    source_and_destination_pairs = source_and_destination_pairs.collect()
    if not source_and_destination_pairs or len(source_and_destination_pairs) == 0:
        if sources is not None or destinations is not None:
            raise ValueError(f"Could not find any pair of source and destination on the graph. Sources {sources} - Destinations {destinations}.")
        else:
            raise ValueError(f"Could not find any pair of source and destination on the graph.")

    paths_df = get_sedona_context().createDataFrame(
        (_compute_shortest_path(graph, graph_costs,
                               row[flows_origin_id_column], row[flows_destination_id_column],
                               row['origin_node'], row['destination_node']
                               ) for row in source_and_destination_pairs),
        [flows_origin_id_column, flows_destination_id_column, 'path', 'geometry', 'path_cost'] +
        [cost.name for cost in graph_costs]
    )
    flows_df = (flows_df
                .join(paths_df, [flows_origin_id_column, flows_destination_id_column], how='inner')
                .withColumn('path_cost', func.col('path_cost') * func.col(flows_demand_column))
                )

    if flows_df_type is pd.DataFrame:
        return gpd.GeoDataFrame(flows_df.toPandas(), geometry = 'geometry', crs = DEFAULT_EPSG.value)
    else:
        return flows_df


#!/usr/bin/env python3
import highspy
import numpy as np
import logging
from ..models import AdalinaData, AdalinaModelType, AdalinaSolution

class AdalinaModelHighs:

    def __init__(self, data : AdalinaData,
                 modeltype : AdalinaModelType = AdalinaModelType.MINSUM_ALLCOST,
                 ub_unserved = None,
                 ub_sum_assigned_distance = None,
                 log_fout = None):

        self.usagevar = None
        self.constr_UB_rhomax = None
        self.model_status = None
        self.log_fout = log_fout

        assert isinstance(data, AdalinaData)
        assert isinstance(modeltype, AdalinaModelType)

        self.modeltype = modeltype

        self.solution = None
        self.run_status = None
        self.objective_value = None

        self.rhomaxvar = None
        self.data = data

        self.vardict = {}
        self.varindex = {}
        self.dict_rownames = {}
        self.count_deletedrows = 0
        self.model = highspy.Highs()
        inf = highspy.kHighsInf

        ###
        ### VARIABLES
        ###

        logging.debug("INIT VARIABLES")

        logging.debug("x")

        # x[s,a] in [0,1]
        self.xvar = []
        for s in self.data.S:
            _xvar = []

            if not self.data.get_shelter_usable(s):
                continue

            for a in self.data.A:

                if not data.get_is_reachable(s, a):
                    continue

                _xvar.append(
                    self._add_variable(0, 1,
                        self._set_xvar_objfun_coeff(s,a),
                        highspy.HighsVarType.kContinuous,
                   f"x_{s}_{a}")
                )
            self.xvar.append(_xvar)

        logging.debug("rho")

        # rho[s,s,r] in [0,1]
        self.rhovar = []
        for s in self.data.S:
            _rhovar = []
            for s1 in self.data.S:

                if not self.data.get_transport_possible(s, s1):
                    continue

                _rhovarS = []
                for r in self.data.R:
                    _rhovarS.append(
                        self._add_variable(0,1,
                                           self._set_rhovar_objfun_coeff(s,s1,r),
                                           highspy.HighsVarType.kContinuous,
                                           f"rho_{s}_{s1}_{r}"
                                           )
                    )
                _rhovar.append(_rhovarS)
            self.rhovar.append(_rhovar)

        logging.debug("y")

        # y[s] in {0,1}
        self.yvar = []
        for s in self.data.S:
            self.yvar.append(
                self._add_variable(0,1,
                                   self._set_yvar_objfun_coeff(s),
                                   # self.data.get_fixed_cost_shelter(s),
                                   highspy.HighsVarType.kInteger,
                                 f"y_{s}")
                             )

        logging.debug("w")

        # w[a] in [0,1]
        self.wvar = []
        for a in self.data.A:
            self.wvar.append(
                self._add_variable(0,1,
                                   self._set_wvar_objfun_coeff(a),
                                   #self.data.get_cost_notserved(a)*self.data.get_area_demand(a),
                                   highspy.HighsVarType.kContinuous,
                                   name=f"w_{a}")
            )

        logging.debug("z")

        # z[s,r] in R >= 0
        self.zvar = []
        for s in self.data.S:
            _zvar = []
            for r in self.data.R:
                _zvar.append(
                    self._add_variable(0, inf,
                                       self._set_zvar_objfun_coeff(s),
                                       # self.data.get_cost_exceed_cap(s),
                                       highspy.HighsVarType.kContinuous,
                                       f"z_{s}_{r}")
                )
            self.zvar.append(_zvar)

        ###
        ### CONSTRAINTS
        ###

        logging.debug("INIT CONSTRAINTS")

        logging.debug("demand cover")

        # forall a in A:  sum{s in S} r[s,a] x[s,a] + w[a] = 1
        # self.constr_cover = []
        for i, a in enumerate(self.data.A):

            _varlist = [self._get_vardict_index(f'x_{s}_{a}')
                            for s in self.data.S if self.data.get_is_reachable(s, a)]

            self._my_add_row(1.0,1.0,
                              len(_varlist)+1,
                              _varlist +
                                [ self._get_vardict_index(f'w_{a}') ],
                              np.array([1]*(len(_varlist)+1)),
                              name = f"res_usage_{i}"
                              )

        logging.debug("resource usage")

        # forall s in S, r in S: n[r] sum{a in A} d[a]x[s,a] - sum{s1 in S} rho[s1, s, r] * o[s1,r] - z[s,r] <= 0
        # self.constr_maxcap = []
        for i,s in enumerate(self.data.S):
            # self.constr_maxcap.append([])
            for j, r in enumerate(self.data.R):

                _vars = []
                _varvalues = []
                for a in self.data.A:

                    if not self.data.get_is_reachable(s, a):
                        continue

                    val = self.data.get_area_demand(a) * self.data.get_resource_need(r)
                    if val > 1e-5:
                        _vars.append(self._get_vardict_index(f'x_{s}_{a}'))
                        _varvalues.append(val)

                for s1 in self.data.S:
                    if self.data.get_transport_possible(s1, s) and self.data.get_shelter_usable(s):

                        _val = data.get_resource_available(s1, r)

                        if _val > 1e-5:
                            _varvalues.append(-data.get_resource_available(s1, r))
                            _vars.append(self._get_vardict_index(f'rho_{s1}_{s}_{r}'))

                self._my_add_row(-inf, 0,
                              len(_vars) + 1,
                              _vars +
                               [ self._get_vardict_index(f'z_{s}_{r}') ],
                              np.array(_varvalues + [-1]),
                            f"usage_res_f{i}_r{j}"
                   )

        logging.debug("not exceeding initial resources")

        # forall s in S, r in R: sum[s1 in S] rho[s,s1,r] <= 1
        for i, s in enumerate(self.data.S):
            for j, r in enumerate(self.data.R):
                _vars = []
                for s1 in self.data.S:
                    varname = f'rho_{s}_{s1}_{r}'
                    if not self._check_vardict_exists(varname):
                        continue
                    _vars.append(self._get_vardict_index(f'rho_{s}_{s1}_{r}'))

                if len(_vars) == 0:
                    continue

                self._my_add_row(-inf, 1,
                                      len(_vars),
                                      _vars,
                                      np.ones(len(_vars), dtype=int),
                                 f"max_usage_f{s}_r{r}"
                )


        logging.debug("link z and y")

        for i, s in enumerate(self.data.S):

            if self.data.get_shelter_usable(s):

                #self.constr_link_zy.append(
                for j, r in enumerate(self.data.R):
                    self._my_add_row(-inf, 0,
                                          2,
                                          np.array([self._get_vardict_index(f'z_{s}_{r}'),
                                           self._get_vardict_index(f'y_{s}')
                                             ]),
                                            np.array(
                                                [1,
                                                 -self.data.get_bigMdemand()
                                                 ]
                                            ),
                                     f"link_zy_f{i}_r{j}"
                                          )

        logging.debug("forbidden assignments")

        # forall s in S, a in A: x[s,a] <= r[a,s]
        # for s in self.data.S:
        #     for a in self.data.A:
        #         if not self.data.get_is_reachable(s, a):
        #             self._set_variable_to_value(f'x_{s}_{a}', 0)

        logging.debug("link rho y, forbidden relocation")

        # forall s, s1 in S, r in R:   rho[s,s1,r] - r[s,s1] bar y[s1] y[s1] <= 0
        #self.constr_no_transport_unreachable = []
        #self.constr_link_rhoy = []
        for i, s in enumerate(self.data.S):
            #self.constr_no_transport_unreachable.append([])
            #self.constr_link_rhoy.append([])
            for j, s1 in enumerate(self.data.S):

                if not self.data.get_transport_possible(s, s1) or not self.data.get_shelter_usable(s1):

                    #self.constr_no_transport_unreachable[i].append([
                    for r in self.data.R:
                        varname = f'rho_{s}_{s1}_{r}'
                        if self._check_vardict_exists(varname):

                            self._set_variable_to_value(varname, 0)

                else:
                    #self.constr_link_rhoy[i].append([
                    for w, r in enumerate(self.data.R):
                        varname = f'rho_{s}_{s1}_{r}'
                        if self._check_vardict_exists(varname):
                            self._my_add_row(-inf, 0,
                                              2,
                                              [self._get_vardict_index(f'rho_{s}_{s1}_{r}'),
                                               self._get_vardict_index(f'y_{s1}')
                                               ],
                                              np.array([1, -1]),
                                             f"link_rho_y_from{i}_to{j}_r{w}"
                                              )


        if ub_unserved is not None:
            self.set_constr_ub_sumunserved(ub_unserved)

        if ub_sum_assigned_distance is not None:
            self.data.ub_total_assignment_distance = ub_sum_assigned_distance
            self.set_constr_ub_sumdist()

        logging.debug("INIT OBJECTIVE")

        if self.modeltype == AdalinaModelType.MIN_EXTRARESOURCES:
            # forall a in A: w[a] = 0
            self._set_w_to_zero()

        if self.modeltype == AdalinaModelType.MINSUM_UNSERVED:

            # set to 0 obj. coefficient of all variables except w[a]
            # self._reset_all_objcoeff(["w_"])

            self._set_z_to_zero()

        if self.modeltype == AdalinaModelType.MINMAX_UNSERVED:

            self._set_z_to_zero()

            # set to 0 obj. coefficient of all variables
            # self._reset_all_objcoeff()

            # add max w variable, only variable in the objective
            self.Wmaxvar = self._add_variable(0, +inf, 1,
                               highspy.HighsVarType.kContinuous, "Wmax")

            # add constraint Wmax >= d[a]w[a] forall a in A
            for i,a in enumerate(self.data.A):
                self._my_add_row(0, inf, 2,
                                  [self._get_vardict_index('Wmax'),
                                   self._get_vardict_index(f'w_{a}')],
                                  np.array([1, -self.data.get_area_demand(a)]),
                                 f"val_wmax_a{i}"
                                 )

        if self.modeltype == AdalinaModelType.MINMAX_USAGEFAC:
            self._set_z_to_zero()
            self._set_w_to_zero()

            self.set_model_minmax_rho()

        ###
        ### OBJECTIVE FUNCTION
        ###

        self.model.setMinimize()

    def _set_variable_to_value(self, varname, lbvalue, ubvalue=None):
        vari = self._get_vardict_index(varname)

        if ubvalue is None:
            ubvalue = lbvalue
        self.model.changeColBounds(vari, lbvalue, ubvalue)

    def _my_add_row(self, lb, ub, numvar, variables, varvalues, name):

        assert isinstance(name, str)

        self.model.addRow(lb, ub, numvar, variables, varvalues)
        self.model.passRowName(len(self.dict_rownames), name)
        self.dict_rownames[len(self.dict_rownames)] = name

    def reset_solution(self):
        self.solution = None
        self.run_status = None
        self.objective_value = None

    def set_model_minsum_unserved(self):
        self._set_z_to_zero()
        self.modeltype = AdalinaModelType.MINSUM_UNSERVED

        for a in self.data.A:
            i = self._get_vardict_index(f"w_{a}")
            self.model.changeColCost(i, self._set_wvar_objfun_coeff(a))

    def set_model_minmax_rho(self):

        inf = highspy.kHighsInf

        # add rho max variable, the only variable in the objective
        self.rhomaxvar = self._add_variable(0, +inf, 1,
                                            highspy.HighsVarType.kContinuous, "rhomax")

        # add constraint Wmax >= (sum{s1 in S, r in R} rho[s1,s,r]*o0[s1,r]) / sum{r in R} o0[s, r]   forall s in S

        logging.debug("constraint rho max")
        # self.constr_rhomax = []
        for i, s in enumerate(self.data.S):

            if not self.data.get_shelter_usable(s):
                continue

            _vars = [self._get_vardict_index('rhomax')]

            _val = self.data.resources_available.xs(s, level=0)["maximum"].sum()
            if _val < 1e-5:
                continue

            _varvalues = [_val]

            for r in self.data.R:
                for s1 in self.data.S:

                    varname = f'rho_{s1}_{s}_{r}'
                    if not self._check_vardict_exists(varname):
                        continue

                    _val = self.data.get_resource_available(s1, r)

                    if _val < 1e-5:
                        continue

                    _vars.append(self._get_vardict_index(varname))
                    _varvalues.append(-self.data.get_resource_available(s1, r))

            self._my_add_row(0, inf,
                              len(_vars),
                              _vars,
                              np.array(_varvalues),
                              f"link_rhomax_fac_{i}"
                              )


    def set_constr_ub_sumunserved(self, ub_unserved):
        logging.debug("constraint upper bound on amount of unserved demand")
        self._my_add_row(-highspy.kHighsInf, ub_unserved,
                              self.data.A_size,
                              [self._get_vardict_index(f'w_{a}') for a in self.data.A],
                              np.array([self.data.get_area_demand(a) for a in self.data.A]),
                         "UB_sum_unserved"
                              )

    def set_constr_ub_sumdist(self):

        logging.debug("constraint upper bound on amount of total transportation distance")

        _vars = []
        _varvalues = []
        for s in self.data.S:
            for a in self.data.A:

                _val = self.data.get_cost_assign(s, a) * self.data.get_area_demand(a)

                if _val > 1e-5:
                    _vars.append(self._get_vardict_index(f'x_{s}_{a}'))
                    _varvalues.append(self.data.get_cost_assign(s, a) * self.data.get_area_demand(a))

        # self.constr_ub_sum_assigned_distance = \
        self._my_add_row(-highspy.kHighsInf, self.data.ub_total_assignment_distance,
                              len(_vars),
                              _vars,
                              np.array(_varvalues),
                         "UB_sum_assign_dist")

    def reset_wvar_costs(self):
        for a in self.data.A:
            i = self._get_vardict_index(f"w_{a}")
            self.model.changeColCost(i, 0)

    def modify_model_to_minsumdist(self):

        # self.reset_solution()

        self.modeltype = AdalinaModelType.MINDIST_ASSIGNMENTS

        for a in self.data.A:
            for s in self.data.S:
                j = self._get_vardict_index(f"x_{s}_{a}")
                self.model.changeColCost(j,
                         self.data.get_cost_assign(s, a) * self.data.get_area_demand(a)
                )

    def modify_model_to_maxdist_unserved(self):

        self.modeltype = AdalinaModelType.MAXSUM_DIST_UNSERVED

        for a in self.data.A:
            for s in self.data.S:

                _ub = 1

                if not self.data.get_is_reachable(s, a):
                    _ub = 0

                self._add_variable(0,_ub,
                                   self.data.get_area_demand(a) * self.data.get_cost_assign(s,a),
                                   highspy.HighsVarType.kContinuous, f"w2_{a}_{s}")

        # sum{s in S} w[a,s] = w[a]
        for i, a in enumerate(self.data.A):
            _vars = [self._get_vardict_index(f"w_{a}")]
            _varvalues = [-1]

            for s in self.data.S:
                _vars.append(self._get_vardict_index(f"w2_{a}_{s}"))
                _varvalues.append(1)

            self._my_add_row(0,0,
                             len(_vars),
                             _vars,
                             _varvalues,
                             f"link_wa_was_{i}")

        # self.model.maximize()

    def reset_x_val_cost(self):
        # set to 0 obj. fun. coefficient of variables x
        for a in self.data.A:
            for s in self.data.S:
                j = self._get_vardict_index(f"x_{s}_{a}")
                self.model.changeColCost(j,0)

    def modify_model_to_minmax_usage_shelter(self):

        self.modeltype = AdalinaModelType.MINMAX_USAGEFAC
        self.set_model_minmax_rho()

    def reset_rhomax_var_objfun_coeff(self):
        rhoindex = self._get_vardict_index("rhomax")
        self.model.changeColCost(rhoindex, 0)

    def set_constr_ub_maxusage(self, value):

        self.data.ub_max_usage = value
        self._set_variable_to_value("rhomax", value)

    def modify_model_to_minsum_usage_shelter(self):

        self.modeltype = AdalinaModelType.MINSUM_USAGEFAC

        logging.debug("constraint rho max")
        # self.constr_UB_rhomax = []

        self.usagevar = []

        for i, s in enumerate(self.data.S):

            if not self.data.get_shelter_usable(s):
                continue

            _total_res = self.data.resources_available.xs(s, level=0)["maximum"].sum()
            if _total_res < 1e-5:
                continue

            self.usagevar.append(self._add_variable(0, self.data.ub_max_usage, 1,
                                                highspy.HighsVarType.kContinuous,
                                                     f"usagevar_{s}")
                                  )

            _vars = [self._get_vardict_index(f"usagevar_{s}")]
            _varvalues = [-_total_res]

            for r in self.data.R:
                for s1 in self.data.S:

                    _val = self.data.get_resource_available(s1, r)

                    if _val < 1e-5:
                        continue

                    _varname = f'rho_{s1}_{s}_{r}'

                    if not self._check_vardict_exists(_varname):
                        continue

                    _index = self._get_vardict_index(_varname)

                    # * self.data.get_cost_resource_transport(s1, s, r))

                    _vars.append(_index)
                    _varvalues.append(self.data.get_resource_available(s1, r))


            self._my_add_row(-highspy.kHighsInf, 0,
                              len(_vars),
                              _vars,
                              np.array(_varvalues),
                             f"UB_usage_f{i}"
                              )

    def set_UB_minsum_usage(self):
        _vars = []
        _varvalues = []
        for s in self.data.S:

            if not self.data.get_shelter_usable(s):
                continue

            _total_res = self.data.resources_available.xs(s, level=0)["maximum"].sum()
            if _total_res < 1e-5:
                continue

            _i = self._get_vardict_index(f"usagevar_{s}")

            self.model.changeColCost(_i, 0)

            _vars.append(_i)
            _varvalues.append(1)

        self._my_add_row(-highspy.kHighsInf, self.data.ub_sum_usage,
                          len(_vars), _vars, np.array(_varvalues),
                         "UB_sum_usage")

    def modify_model_to_minsum_relocation_costs(self):

        self.modeltype = AdalinaModelType.MINSUM_RELOCATION

        for s in self.data.S:
            for s1 in self.data.S:
                if s == s1:
                    continue
                for r in self.data.R:

                    _varname = f"rho_{s}_{s1}_{r}"

                    if not self._check_vardict_exists(_varname):
                        continue

                    self.model.changeColCost(
                        self._get_vardict_index(_varname),
                        self._set_rhovar_objfun_coeff(s, s1, r)
                    )

    def _set_zvar_objfun_coeff(self, shelter):

        if self.modeltype == AdalinaModelType.MIN_EXTRARESOURCES:
            return 1

        if self.modeltype != AdalinaModelType.MINSUM_ALLCOST:
            return 0

        return self.data.get_cost_exceed_cap(shelter)

    def _set_wvar_objfun_coeff(self, area):

        if self.modeltype == AdalinaModelType.MINSUM_UNSERVED:
            return self.data.get_area_demand(area)

        if self.modeltype != AdalinaModelType.MINSUM_ALLCOST:
            return 0

        return self.data.get_cost_notserved(area) * self.data.get_area_demand(area)

    def _set_yvar_objfun_coeff(self, shelter):

        if self.modeltype != AdalinaModelType.MINSUM_ALLCOST:
            return 0

        return self.data.get_fixed_cost_shelter(shelter)

    def _set_xvar_objfun_coeff(self, shelter, area ):

        #if self.modeltype in [AdalinaModelType.MINDIST_ASSIGNMENTS, AdalinaModelType.MINSUM_ALLCOST] :
        return self.data.get_demcost_assign(shelter,area, self.modeltype)# *self.data.get_area_demand(area)

        # return 0

    def _set_rhovar_objfun_coeff(self, shelter_from, shelter_to, resources):

        return self.data.get_cost_resource_transport(shelter_from, shelter_to, resources, self.modeltype)

    def reset_all_objcoeff(self, noreset_prefix = None):
        # set to 0 obj. coefficient of all variables
        for i, var in enumerate(self.model.getVariables()):
            if noreset_prefix is not None and sum([var.name.startswith(el) for el in noreset_prefix]) > 0:
                continue
            self.model.changeColCost(i, 0)

    def _set_w_to_zero(self):
        # forall a in A: w[a] = 0
        for a in self.data.A:
            self._set_variable_to_value(f'w_{a}', 0)

    def _set_z_to_zero(self):
        # set to 0 variable z
        for s in self.data.S:
            for r in self.data.R:
                self._set_variable_to_value(f'z_{s}_{r}', 0)

    def set_rhovar_to_zero(self):

        self.set_rhovar_by_distance_threshold(-1)

    def set_rhovar_by_distance_threshold(self, threshold ):
        for s in self.data.S:
            for s1 in self.data.S:
                if s == s1:
                    continue
                for r in self.data.R:

                    _varname = f"rho_{s}_{s1}_{r}"

                    if not self._check_vardict_exists(_varname):
                        continue

                    _facility_distance = self.data.get_facilities_distance(s, s1)
                    if _facility_distance is None:
                        self._set_variable_to_value(_varname, 0)

                    elif _facility_distance > threshold:
                        self._set_variable_to_value(_varname, 0)
                    else:
                        self._set_variable_to_value(_varname, 0, 1)

    def _add_variable_dict_element(self, name):
        _index = len(self.vardict)
        self.vardict[name] = _index
        self.varindex[_index] = name

        return _index

    def _add_variable(self, lb, ub, obj, vartype, name):
        var = self.model.addVariable(lb, ub,
                                   obj = obj,
                                   type=vartype,
                                   name=name)

        self._add_variable_dict_element(name)

        return  var

    def _check_vardict_exists(self, name):
        return name in self.vardict

    def _get_vardict_index(self, name):
        return self.vardict[name]

    def run(self, timelimit=60):

        logging.debug(f"INIT RUN WITH TIME LIMIT {timelimit} SECONDS")

        self.model.setOptionValue("time_limit",
            #highspy.HighsOptions.time_limit,
                                  timelimit)

        self.run_status = self.model.run()

        if self.run_status == highspy.HighsStatus.kOk:

            self.model_status = self.model.getModelStatus()

            if self.check_model_feasible():
                self.get_solution()
                return True
            else:
                logging.debug("solution not available")
                logging.debug(self.model_status)
                return False
        else:
            return False

    def _build_solution(self):

        logging.debug("building solution")

        self.solution = AdalinaSolution(self.objective_value) # self.model.getObjectiveValue())

        # allvars = self.model.getVariables()

        allvarvalues = np.array(self.model.variableValues(self.model.getVariables()))

        for varindex in np.where(allvarvalues > 1e-5)[0]:
            val = allvarvalues[varindex]
            varname = self.varindex[varindex]

            l = varname.split("_")

            if l[0].startswith('x'):
                self.solution.add_xvarvalue(l[1], l[2], val)
            elif l[0].startswith('y'):
                self.solution.add_yvarvalue(l[1], val)
            elif l[0] == 'w':
                self.solution.add_wvarvalue(l[1], val)
            elif l[0] == 'w2':
                self.solution.add_w2varvalue(l[1], l[2], val)
            elif l[0].startswith('z'):
                self.solution.add_zvarvalue(l[1], l[2], val)
            elif l[0] == 'rho':
                self.solution.add_rhovalue(l[1],l[2],l[3], val)
            elif self.modeltype == AdalinaModelType.MINMAX_UNSERVED:
                self.solution.set_wmaxvalue(val)
            elif self.modeltype == AdalinaModelType.MINMAX_USAGEFAC:
                self.solution.set_rhomaxvalue(val)
            elif l[0] == "usagevar":
                self.solution.add_usagevarvalue(l[1], val)

        if self.modeltype == AdalinaModelType.MINMAX_UNSERVED and not self.solution.check_wmax_set():
            self.solution.set_wmaxvalue(0)
        if self.modeltype == AdalinaModelType.MINMAX_USAGEFAC and not self.solution.check_rhomax_set():
            self.solution.set_rhomaxvalue(0)

        logging.debug("check feasibility of solution")
        return self.solution.check_feasibility(self.data, self.modeltype)

    def check_model_feasible(self):

        if self.model_status is None:
            return False

        return self.model_status in [highspy.HighsModelStatus.kOptimal, highspy.HighsModelStatus.kIterationLimit,
                                     highspy.HighsModelStatus.kTimeLimit
                                     ]

    def get_solution(self, run_model=True):

        if self.solution is not None:
            return self.solution

        elif self.check_model_feasible():

            self.objective_value = self.model.getObjectiveValue()

            if self._build_solution():
                logging.debug("feasibility double check passed!")

                return self.solution

            else:
                logging.debug(f"INFEASIBILITIES WITH POST-PROCESSING CONTROL: {self.solution.infeasibility_log}")
            # print_log("model infeasible. no solution")
            return None

        if self.solution is None and run_model:

            if self.run() == highspy.HighsStatus.kOk:
                if self._build_solution():
                    logging.debug("feasibility double check passed!")
                elif self.solution is not None:
                    logging.debug(f"INFEASIBILITIES WITH POST-PROCESSING CONTROL: {self.solution.infeasibility_log}")

                return self.solution

            return None

        elif self.solution is not None:
            return self.solution

        logging.debug("no solution available. run model with function run()")
        return None




from pydantic import BaseModel, Field, field_validator

_zoning_functions = ["euclidean", "manhattan", "chebyshev", "haversine", "hamming", "canberra", "braycurtis", "jaccard", "matching", "dice", "kulsinski", "rogerstanimoto", "russellrao", "sokalmichener", "sokalsneath"]

class ClusteringDistance(BaseModel):
    name: str = Field(description='Name of the column')
    """Name of the column in the input data to be used for distance calculation."""

    weight: int | None = Field(description='Weight of the column', default=1)
    """Weight of the distance in the overall distance calculation. Default is 1."""

    function: str | None = Field(description='Function used to compute the distance', default=None)
    """
    Distance function to use. Must be one of the following:
        - euclidean
        - manhattan
        - chebyshev
        - haversine
        - hamming
        - canberra
        - braycurtis
        - jaccard
        - matching
        - dice
        - kulsinski
        - rogerstanimoto
        - russellrao
        - sokalmichener
        - sokalsneath
    If None, the default function depends on the type of the values in the column:
        - geometric values default to scipy distance_matrix
        - float and int values default to euclidean
        - string values default to hamming
        - boolean values default to jaccard
    """

    @field_validator('function')
    def validate_func(cls, v):
        if v is not None and v not in _zoning_functions:
            raise ValueError(f'Function must be one of {_zoning_functions}')
        return v


import pytest
from libadalina_analytics.graph_extraction.readers import OpenStreetMapReader
import geopandas as gpd
import pathlib

from libadalina_analytics.graph_extraction.readers import MandatoryColumns

SAMPLE_DIR = pathlib.Path(__file__).parent.parent / "samples"

@pytest.fixture
def osm_reader():
    return OpenStreetMapReader()

class TestOpenStreetMapReader:

    @pytest.mark.parametrize("input_map", [
        f"{SAMPLE_DIR}/road_maps/Milano.gpkg",
        f"{SAMPLE_DIR}/road_maps/Milano.csv",
    ])
    def test_read(self, osm_reader, input_map):
        """Test reading OpenStreetMap data."""
        osm_data = osm_reader.read(input_map)
        assert osm_data is not None
        assert isinstance(osm_data, gpd.GeoDataFrame)
        assert len(osm_data) > 0
        assert all(c.value in osm_data.columns for c in MandatoryColumns)

    def test_read_with_invalid_file(self, osm_reader):
        """Test reading with an invalid file."""
        with pytest.raises(FileNotFoundError):
            osm_reader.read("invalid_file.csv")